import datetime, json, logging, os

import boto3

from . import credentials as cred


logger = logging.getLogger(__name__)


def get_client(cred_keys: dict = {}):
    return cred.get_client("logs", cred_keys)


def get_log_group_arn(log_group_name, cred_keys: dict = {}):
    client = get_client(cred_keys)
    info = client.describe_log_groups(
            logGroupNamePrefix=log_group_name,
            limit=1,
            )
    return info["logGroups"][0]["arn"]


def filter_log_events(
    log_group_name,
    start_time=None,
    end_time=None,
    next_token=None,
    filter_pattern=None,
    extras=None,
    cred_keys: dict = {},
):
    if not filter_pattern:
        filter_pattern = " "

    client = get_client(cred_keys)
    kwargs = {
        "logGroupName": log_group_name,
        "filterPattern": filter_pattern,
    }
    if start_time:
        if type(start_time) is datetime.datetime:
            kwargs["startTime"] = int(start_time.timestamp() * 1000)
        elif type(start_time) is int:
            kwargs["startTime"] = start_time
        else:
            raise ValueError("Invalid startTime type")

    if end_time:
        if type(end_time) is datetime.datetime:
            kwargs["endTime"] = int(end_time.timestamp() * 1000)
        elif type(end_time) is int:
            kwargs["endTime"] = start_time
        else:
            raise ValueError("Invalid endTime type")

    if next_token:
        kwargs["nextToken"] = next_token

    valid_params = [
            "logGroupName",
            "logStreamNames",
            "logStreamNamePrefix",
            "startTime",
            "endTime",
            "filterPattern",
            "nextToken",
            "limit",
            "interleaved",
            ]
    if extras:
        for k in extras.keys():
            if k in valid_params:
                kwargs[k] = extras[k]

    response = client.filter_log_events(**kwargs)
    return response



#!/usr/bin/python
import io, json, logging, sys, os

LOGGER = logging.getLogger()

def collect_stats(outfile):
    """
    Read from stdin and count the lines. Output to a file after done.
    """
    lines = io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8")
    output = {"rows_read": 0}
    for line in lines:
        try:
            o = json.loads(line)
            print(json.dumps(o))
            if o["type"].lower() == "record":
                output["rows_read"] += 1
        except json.decoder.JSONDecodeError:
            print(line)
            output["rows_read"] += 1
    with open(outfile, "w") as f:
        json.dump(output, f)
        f.write("\n")


if __name__ == "__main__":
    collect_stats("artifacts/collect_stats.json")



import datetime, json, logging, os, re, sys
from dateutil.parser import parse as dateparse


LOGGER = None


def get_logger(name="handoff"):
    global LOGGER
    if not LOGGER:
        logging.basicConfig(
            stream=sys.stdout,
            format="[%(asctime)s] [%(levelname)8s] - %(message)s - (%(filename)s:%(lineno)s)",
            level=logging.INFO)
        LOGGER = logging.getLogger(name)
    return LOGGER


def increment_version(ver, delimiter="."):
    """Increment the last digit of the version numbers with delimiter "."
    if it fails to increment, returns original version by adding ".1"
    """
    if ver is None:
        return "0.1"
    digits = ver.split(delimiter)
    for i in range(len(digits) - 1, 0, -1):
        try:
            digit = int(digits[i])
        except ValueError:
            continue
        digits[i] = str(digit + 1)
        new_version = ".".join(digits)
        return new_version
    return (ver + ".1")


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def time_trunc(org_datetime, by='minute'):
    """Convenience function to truncate the timestamp by the specified unit"""
    if by not in ['year', 'month', 'day', 'hour', 'minute', 'second']:
        raise ValueError('Invalid time unit for truncation')
    t_elems = {'year': org_datetime.year, 'month': org_datetime.month,
               'day': org_datetime.day,
               'hour': org_datetime.hour, 'minute': org_datetime.minute,
               'second': org_datetime.second,
               'microsecond': org_datetime.microsecond}
    # Fix for Python 3.x: It does not honor the original key order of the dict
    keys = ["year", "month", "day", "hour", "minute", "second", "microsecond"]
    found = False
    for k in keys:
        if found:
            t_elems[k] = 1 if k in ['month', 'day'] else 0
        if k == by:
            found = True
    return datetime.datetime(**t_elems)


def get_time_window(data,
                    date_trunc="day",
                    start_offset=datetime.timedelta(days=0),
                    end_offset=datetime.timedelta(days=1),
                    iso_str=True):
    """
    Check the input data for start_at and end_at.
    If not present, determine them by relative time from now.
    Returns the ISO formatted strings in UTC.

    - data: A dictionary that may or may not contain "start_at" and "end_at"
    - date_trunc: "day", "hour", "minute" to truncate the time.
    - start_offset: Usually a negative timedelta to go back in time from now.
    - end_offset: Timedelta from start_at to determine the window. The exact
      end_at is not included in the data point.
    """
    start_at = data.get("start_at")
    end_at = data.get("end_at")

    if type(start_at) == str:
        start_at = dateparse(start_at)
    if type(end_at) == str:
        end_at = dateparse(end_at)

    if type(start_at) != datetime.datetime:
        # Time-truncated version of now
        now = time_trunc(datetime.datetime.utcnow(), date_trunc)
        start_at = now + start_offset

    if type(end_at) != datetime.datetime:
        end_at = start_at + end_offset

    if (type(start_at) == datetime.datetime and
            type(end_at) == datetime.datetime):
        if start_at >= end_at:
            raise ValueError("start_at must be strictly earlier than end_at")
        if iso_str:
            return start_at.isoformat(), end_at.isoformat()
        return start_at, end_at

    raise ValueError("Something went wrong determining start_at and end_at")



import os, tarfile
from importlib import import_module as _import_module
from typing import Dict

import yaml

from handoff.utils import get_logger as _get_logger
from handoff.config import get_state, BUNDLES_DIR, CONTAINER_PROVIDER, TASK_NAKED
from handoff.services import cloud


CONTAINER_MODULE = None


def _get_platform(provider_name: str = None) -> str:
    state = get_state()
    if not provider_name:
        provider_name = state.get(CONTAINER_PROVIDER)
    global CONTAINER_MODULE
    if not CONTAINER_MODULE:
        if not provider_name:
            raise Exception(
                "You need to set container provider name (e.g. docker)")
        CONTAINER_MODULE = _import_module("handoff.services.container." +
                                          provider_name)
    return CONTAINER_MODULE


def _get_latest_image_version(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> str:
    return _get_platform().get_latest_image_version(
        project_dir, workspace_dir, **kwargs)


def _tar_filter(file_):
    if ".secret" in file_.name:
        return None
    return file_


def bundle(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    build_dir = vars.get("build_dir")
    tarball_dir = vars.get("tarball_dir")
    if not build_dir or not tarball_dir:
        raise Exception("build_dir and tarball_dir must be given")

    platform = cloud.get_platform()
    state = get_state()
    state.validate_env()

    res = _get_platform().bundle(project_dir, workspace_dir, vars, **kwargs)
    if res["status"] != "success":
        raise Exception("Bundle creation failed.")

    platform.delete_dir(BUNDLES_DIR)

    current_dir = os.getcwd()
    tar_file_name = state[TASK_NAKED] + ".tar.gz"
    with tarfile.open(os.path.join(tarball_dir, tar_file_name), "w:gz") as tar:
        os.chdir(build_dir)
        tar.add(
            ".",
            arcname=".",
            recursive=True,
            filter=_tar_filter,
        )
    os.chdir(current_dir)
    remote_tarball_path = f"{BUNDLES_DIR}/{tar_file_name}"
    res = platform.upload_file(
            os.path.join(tarball_dir, tar_file_name),
            remote_tarball_path,
    )


def build(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff container build -p <project_directory> -v docker_file=<docker_file> files_dir=<addtnl_files_dir>`
    Build the container image
    Optionally, use docker_file to specify own Dockerfile
    """
    return _get_platform().build(project_dir, workspace_dir, **kwargs)


def push(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff container push -p <project_directory>`
    Push the container image to remote repository
    """
    return _get_platform().push(project_dir, workspace_dir, **kwargs)


def run(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff container run -p <project_directory> -e resource_group=<resource_group_name> task=<task_name> __VARS='key1=val1 key2=val2...'`
    Run the container

    If the environment variable __VARS is specified via -e option, it will be
    used as:
    `handoff run -d $(eval echo $__VARS)`
    """
    return _get_platform().run(project_dir, workspace_dir, **kwargs)



import datetime, logging

import boto3

from . import credentials as cred
from . import cloudformation as cfn, sts


logger = logging.getLogger(__name__)


def get_client(cred_keys: dict = {}):
    return cred.get_client("ecs", cred_keys)


def describe_tasks(
        resource_group,
        region,
        running=True,
        stopped=True,
        extras=None,
        cred_keys: dict = {},
        ):
    client = get_client(cred_keys=cred_keys)
    account_id = sts.get_account_id(cred_keys=cred_keys)
    cluster = f"arn:aws:ecs:{region}:{account_id}:cluster/{resource_group}"
    task_arns = []
    if stopped:
        response = client.list_tasks(cluster=cluster, desiredStatus="STOPPED")
        task_arns = task_arns + response["taskArns"]
    if running:
        response = client.list_tasks(cluster=cluster, desiredStatus="RUNNING")
        task_arns = task_arns + response["taskArns"]
    tasks = [t for t in task_arns]
    if not tasks:
        return None
    response = client.describe_tasks(cluster=cluster, tasks=tasks)
    return response


def stop_task(resource_group, region, task_id, reason, extras=None,
        cred_keys: dict = {}):
    client = get_client(cred_keys=cred_keys)
    account_id = sts.get_account_id(cred_keys=cred_keys)
    cluster = f"arn:aws:ecs:{region}:{account_id}:cluster/{resource_group}"
    response = client.stop_task(cluster=cluster, task=task_id, reason=reason)
    return response


def run_fargate_task(
        account_id,
        task_name,
        resource_group_stack,
        container_name,
        region,
        env=[],
        extras=None,
        command=None,
        cred_keys: dict = {}):
    """Run a fargate task
    extras overwrite the kwargs given to run_task boto3 command.
    See: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html#ECS.Client.run_task
    """
    env.append({"name": "TASK_TRIGGERED_AT",
                "value": datetime.datetime.utcnow().isoformat()})
    client = get_client(cred_keys=cred_keys)

    rg_resources = cfn.describe_stack_resources(
        resource_group_stack,
        cred_keys=cred_keys,
    )["StackResources"]

    rg_resources = cfn.describe_stack_resources(
        resource_group_stack,
        cred_keys=cred_keys,
    )["StackResources"]
    cluster_arn = None
    subnets = list()
    security_groups =list()
    for r in rg_resources:
        # PublicSubnetTwo is for <= v0.3.4
        if (r["ResourceType"] == "AWS::EC2::Subnet" and
                r["LogicalResourceId"] in ["PublicSubnetTwo", "Subnet2"]):
            subnets.append(r["PhysicalResourceId"])
        if r["ResourceType"] == "AWS::EC2::SecurityGroup":
            security_groups.append(r["PhysicalResourceId"])
        if r["ResourceType"] == "AWS::ECS::Cluster":
            cluster_arn = "arn:aws:ecs:{region}:{account_id}:cluster/{phys_rsrc_id}".format(
                **{"region": region,
                   "account_id": account_id,
                   "phys_rsrc_id": r["PhysicalResourceId"]})

    logger.info(task_name)
#     task_resources = cfn.describe_stack_resources(
#         task_name,
#         cred_keys=cred_keys,
#     )["StackResources"]
#     task_def_arn = None
#     for r in task_resources:
#         if r["ResourceType"] == "AWS::ECS::TaskDefinition":
#             task_def_arn = r["PhysicalResourceId"]
#             break
    task_def_arn = client.list_task_definitions(
        familyPrefix=task_name,
        status="ACTIVE",
        sort="DESC",
        maxResults=1,
    )["taskDefinitionArns"][0]

    logger.debug(
        "%s \n%s \n%s \n%s" % (cluster_arn, task_def_arn, subnets, security_groups)
    )

    container_override = {
        "name": container_name,
        "environment": env,
    }
    if command:
        container_override["command"] = command

    kwargs = {
        "cluster": cluster_arn,
        "taskDefinition": task_def_arn,
        "count": 1,
        "launchType": "FARGATE",
        "networkConfiguration": {
            "awsvpcConfiguration": {
                "subnets": subnets,
                "securityGroups": security_groups,
                "assignPublicIp": "ENABLED"
            }
        },
        "overrides": {
            "containerOverrides": [container_override]
        }
    }
    if extras:
        kwargs.update(extras)

    return client.run_task(**kwargs)



import json, logging, os

import boto3

from . import credentials as cred
from handoff import utils


LOGGER = utils.get_logger()


def get_client(cred_keys: dict = {}):
    return cred.get_client("stepfunctions", cred_keys)


def create_state_machine(
        name,
        cluster_arn,
        task_def_arn,
        subnets,
        security_groups,
        role_arn,
        overrides,
        log_group_arn=None,
        timeout=3600 * 4,
        retry_interval=1800,
        max_retries=4,
        backoff_rate=1,
        cred_keys: dict = {},
        **kwargs):
    client = get_client(cred_keys)
    definition = {
        "Version": "1.0",
        "Comment": "Run ECS/Fargate tasks",
        "TimeoutSeconds": timeout,
        "StartAt": "RunTask",
        "States": {
            "RunTask": {
                "Type": "Task",
                "Resource": "arn:aws:states:::ecs:runTask.sync",
                "Parameters": {
                    "LaunchType": "FARGATE",
                    "Cluster": cluster_arn,
                    "TaskDefinition": task_def_arn,
                    "NetworkConfiguration": {
                        "AwsvpcConfiguration": {
                            "Subnets": subnets,
                            "SecurityGroups": security_groups,
                            "AssignPublicIp": "ENABLED"
                        }
                    },
                    "Overrides": overrides,
                },
                "Retry": [
                    {
                        "ErrorEquals": [
                            "States.TaskFailed"
                        ],
                        "IntervalSeconds": retry_interval,
                        "MaxAttempts": max_retries,
                        "BackoffRate": backoff_rate,
                    }
                ],
                "End": True
            }
        }
    }
    sfn_kwargs = {}
    if log_group_arn:
        sfn_kwargs["loggingConfiguration"] = {
            "level": "ALL",
            "includeExecutionData": False,
            "destinations": [
                {
                    "cloudWatchLogsLogGroup": {
                        "logGroupArn": log_group_arn,
                    }
                },
            ]
        }

    response = client.create_state_machine(
        name=name,
        definition=json.dumps(definition),
        roleArn=role_arn,
        **sfn_kwargs,
    )
    sm_arn = response["stateMachineArn"]
    return sm_arn


def delete_state_machine(
        arn,
        cred_keys: dict = {}):
    client = get_client(cred_keys)
    response = client.delete_state_machine(stateMachineArn=arn)



import logging, os, shutil
import requests
from typing import Dict
import pygit2
from github import Github

from handoff.config import get_state as _get_state
from handoff.config import GITHUB_ACCESS_TOKEN
from handoff.utils import get_logger

LOGGER = get_logger()
GITHUB = None


def _get_github(access_token=None):
    state = _get_state()
    if not access_token:
        access_token = state.get(GITHUB_ACCESS_TOKEN)
    global GITHUB
    GITHUB = Github(access_token)
    return GITHUB


def _get_callbacks(access_token):
    return  pygit2.RemoteCallbacks(pygit2.UserPass(access_token, "x-oauth-basic"))


def token_test(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    state = _get_state()
    access_token = vars.get("access_token", state.get(GITHUB_ACCESS_TOKEN))
    res = requests.get(
        "https://api.github.com",
        headers={"authorization": f"Bearer {access_token}"},
    )
    res = res.json()
    if res.get("authorizations_url"):
        return {"status": "success", "detail": res}
    return {"status": "auth failure", "detail": res}


def branch(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff github branch -v organization=<github_org> repository=<github_repo> local_dir=<local_dir> force=False`
    List local and remote branches. If local_dir is omitted, ./<repository> is used.
    """
    state = _get_state()
    access_token = vars.get("access_token", state.get(GITHUB_ACCESS_TOKEN))
    github = _get_github(access_token)

    local_dir = vars["local_dir"]  # typically the "./{repository_name}"
    branch = vars.get("branch", "master")
    repo = pygit2.Repository(local_dir + "/.git")

    branches = set(repo.branches.local)
    for b in repo.branches.remote:
        branch = b.split("/")[-1]
        branches.add(branch)

    return {
        "status": "success",
        "branches": list(branches),
    }


def clone(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff github clone -v organization=<github_org> repository=<github_repo> local_dir=<local_dir> force=False`
    Clone remote repository. If local_dir is omitted, ./<repository> is used.
    If force is set to True, an existing directory will be deleted before cloning.

    Note: It's actually git init & git pull URL to avoid access token to be written out to .git/config
          See https://github.blog/2012-09-21-easier-builds-and-deployments-using-git-over-https-and-oauth/
    """
    state = _get_state()
    access_token = vars.get("access_token", state.get(GITHUB_ACCESS_TOKEN))
    github = _get_github(access_token)

    repo_name = vars["repository"]
    org_name = vars.get("organization")
    local_dir = vars.get("local_dir", "./" + repo_name)
    if os.path.exists(local_dir):
        if not vars.get("force", False):
            raise Exception("The directory already exists.")
        shutil.rmtree(local_dir)
    git_url = vars.get("url", "https://github.com/" + str(org_name) + "/" + repo_name + ".git")

    if vars.get("use_cli", False):
        LOGGER.debug("Running git CLI")
        git_url = git_url.replace("https://", f"https://{access_token}:x-oauth-basic@")
        git_path = os.environ.get("GIT_PATH", "git")
        os.mkdir(local_dir)
        os.system(f"{git_path} init {local_dir}")
        cur_dir = os.getcwd()
        os.chdir(local_dir)
        os.system(f"{git_path} pull {git_url}")
        os.system(f"{git_path} remote add origin {git_url}")
        os.chdir(cur_dir)
    else:
        repo_clone = pygit2.init_repository(local_dir, True)
        repo_clone.remotes.create("origin", git_url)
        pull(project_dir, workspace_dir, vars, **kwargs)
        # Using clone against GitHub repo with token is a security risk!
        # repo_clone = pygit2.clone_repository(
        #         git_url, local_dir, callbacks=_get_callbacks(access_token))

    return {"status": "success", "repository": repo_name}


def pull(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff github commit -v local_dir=<local_dir> commit_msg=<msg> branch=<branch>`
    Commit all the outstanding changes to the remote branch. If the branch does not exist,
    it will be created.
    """
    state = _get_state()
    access_token = vars.get("access_token", state.get(GITHUB_ACCESS_TOKEN))
    github = _get_github(access_token)

    local_dir = vars["local_dir"]  # typically the "./{repository_name}"
    branch = vars.get("branch", "master")
    repo = pygit2.Repository(local_dir + "/.git")
    remote_name = "origin"

    if vars.get("use_cli", False):
        LOGGER.debug("Running git CLI")
        git_path = os.environ.get("GIT_PATH", "git")
        cur_dir = os.getcwd()
        os.chdir(local_dir)
        os.system(f"{git_path} fetch")
        os.system(f"{git_path} checkout {branch}")
        os.system(f"{git_path} pull")
        os.chdir(cur_dir)
        # TODO: handle merge conflict
        return {
            "status": "success",
            "message": "successfully fast forwarded the repository",
        }

    # PyGit mode
    # Adopted from https://github.com/MichaelBoselowitz/pygit2-examples/blob/68e889e50a592d30ab4105a2e7b9f28fac7324c8/examples.py#L48
    for remote in repo.remotes:
        if remote.name == remote_name:
            remote.fetch(callbacks=_get_callbacks(access_token))
            remote_master_id = repo.lookup_reference(f"refs/remotes/origin/{branch}").target
            merge_result, _ = repo.merge_analysis(remote_master_id)
            # Up to date, do nothing
            if merge_result & pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE:
                return {
                    "status": "success",
                    "message": "Repository is up-to-date",
                }

            # We can just fastforward
            elif merge_result & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD:
                repo.checkout_tree(repo.get(remote_master_id))
                master_ref = repo.lookup_reference(f"refs/heads/{branch}")
                master_ref.set_target(remote_master_id)
                repo.head.set_target(remote_master_id)
            elif merge_result & pygit2.GIT_MERGE_ANALYSIS_NORMAL:
                repo.merge(remote_master_id)

                if repo.index_conflicts:
                    return {
                        "status": "error",
                        "mesage": repo.index.conflicts,
                    }

                user = repo.default_signature
                tree = repo.index.write_tree()
                commit = repo.create_commit('HEAD',
                                            user,
                                            user,
                                            'Merge!',
                                            tree,
                                            [repo.head.target, remote_master_id])
                repo.state_cleanup()
                return {
                    "status": "success",
                    "message": "successfully fast forwarded the repository",
                }
            else:
                return {
                    "status": "error",
                    "mesage": "Unknown merge analysis result"
                }


def commit (
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff github commit -v local_dir=<local_dir> commit_msg=<msg> branch=<branch>`
    Commit all the outstanding changes to the remote branch. If the branch does not exist,
    it will be created.
    """
    state = _get_state()
    access_token = vars.get("access_token", state.get(GITHUB_ACCESS_TOKEN))
    github = _get_github(access_token)

    local_dir = vars["local_dir"]  # typically the "./{repository_name}"
    repo = pygit2.Repository(local_dir + "/.git")

    user = github.get_user()

    branch = vars.get("branch", "master")
    commit_msg = vars["commit_msg"]

    # repo.remotes.set_url("origin", repo.clone_url)
    index = repo.index
    index.add_all()
    index.remove_all("*.secrets*")
    index.write()
    author = pygit2.Signature(user.name, user.email)
    commiter = pygit2.Signature(user.name, user.email)
    tree = index.write_tree()
    oid = repo.create_commit(
            "refs/heads/" + branch,
            author,
            commiter,
            commit_msg,
            tree,
            [repo.head.target],
            # [repo.head.get_object().hex],
            )

    credentials = pygit2.UserPass(access_token, "x-oauth-basic")
    # callbacks = pygit2.RemoteCallbacks(credentials=credentials)
    remote = repo.remotes["origin"]
    remote.credentials = credentials
    remote.push(["refs/heads/" + branch], callbacks=_get_callbacks(access_token))



import json, logging, os, shutil, sys, tempfile
from collections import defaultdict

from packaging import version

import docker
from docker import APIClient as docker_api_client

from handoff import utils
from handoff.config import ADMIN_ENVS, CONTAINER_IMAGE, get_state


logger = utils.get_logger(__name__)

DOCKERFILE = "Dockerfile"


def get_latest_version(image_name, ignore=["latest"]):
    client = docker.from_env()
    images = client.images.list(name=image_name)
    if not images:
        return None

    tags = list()
    max_version = None
    for image in images:
        tags = tags + image.tags
    for tag in tags:
        ver = "".join(tag.split(":")[1:])
        if (ver not in ignore and
                (max_version is None or
                 version.parse(max_version) < version.parse(ver))):
            max_version = ver
    return max_version


def copy_files(project_dir, build_dir, docker_file=None, files_dir=None):
    docker_file_dir, _ = os.path.split(__file__)
    if not docker_file:
        docker_file = os.path.join(docker_file_dir, DOCKERFILE)

    cwd, _ = os.path.split(__file__)
    logger.info("Current working directory: " + cwd)
    handoff_dir = os.path.join(cwd, "../../../../")
    logger.info("Looking for handoff source at " + handoff_dir)
    if os.path.isfile(os.path.join(handoff_dir, "setup.py")):
        logger.info("Found handoff. Copying to the build directory")
        ho_build_dir = os.path.join(build_dir, "handoff")
        os.mkdir(ho_build_dir)
        shutil.copytree(os.path.join(handoff_dir, "handoff"),
                        os.path.join(ho_build_dir, "handoff"))
        files = [
            "MANIFEST.in",
            "README.md",
            "requirements.txt",
            "setup.cfg",
            "setup.py",
        ]
        for fn in files:
            shutil.copyfile(os.path.join(handoff_dir, fn),
                            os.path.join(ho_build_dir, fn))
    else:
        logger.info("...not found")

    shutil.copytree(project_dir, os.path.join(build_dir, "project"),
                    symlinks=True)
    shutil.copytree(os.path.join(docker_file_dir, "script"),
                    os.path.join(build_dir, "script"))
    shutil.copyfile(docker_file, os.path.join(build_dir, DOCKERFILE))

    if files_dir:
        logger.info(f"Copying the directory {files_dir} to {build_dir}")
        path, dir_ = os.path.split(files_dir)
        shutil.copytree(files_dir, os.path.join(build_dir, dir_))
    return {"status": "success"}


def build(project_dir, new_version=None, docker_file=None, files_dir=None,
          nocache=False, yes=False, file_descriptor=sys.stdout, **kwargs):
    state = get_state()
    cli = docker_api_client()
    image_name = state[CONTAINER_IMAGE]
    if not image_name:
        raise Exception(f"You need to set {CONTAINER_IMAGE} environment variable.")

    if not new_version:
        latest_version = get_latest_version(image_name)
        new_version = utils.increment_version(latest_version)

    if not yes:
        sys.stdout.write("Build %s:%s (y/N)? " % (image_name, new_version))
        response = input()
        if response.lower() not in ["y", "yes"]:
            return{"status": "canceled"}

    logger.info("Building %s:%s" % (image_name, new_version))

    with tempfile.TemporaryDirectory() as build_dir:
        copy_files(project_dir, build_dir, docker_file, files_dir)
        for line in cli.build(path=build_dir,
                              tag=f"{image_name}:{new_version}",
                              nocache=nocache):
            for subline in line.decode("utf-8").split("\n"):
                try:
                    msg = json.loads(subline)
                except json.decoder.JSONDecodeError:
                    continue
                if msg.get("stream") and msg["stream"] != "\n":
                    file_descriptor.write(msg["stream"])
        return{"status": "success"}


def run(version=None, extra_env=dict(), yes=False, command=None, detach=False,
        interactive=False, file_descriptor=sys.stdout, **kwargs):
    state = get_state()
    env = {}
    for e in ADMIN_ENVS.keys():
        env[e] = state.get(e)

    env.update(extra_env)

    image_name = state[CONTAINER_IMAGE]
    if not version:
        version = get_latest_version(image_name)

    if not yes:
        sys.stdout.write("Run %s:%s (y/N)? " % (image_name, version))
        response = input()
        if response.lower() not in ["y", "yes"]:
            return

    if interactive:
        stdin_open = True
        tty = True
    else:
        stdin_open = False
        tty = False

    client = docker.from_env()

    try:
        # TODO: Streaming log is not working :(
        for line in client.containers.run(image_name + ":" + version,
                                          command=command, environment=env,
                                          stream=True, detach=detach,
                                          stdin_open=stdin_open, tty=tty):
            file_descriptor.write(line.decode("utf-8").replace("\\n", "\n"))
    except Exception as e:
        return {
            "status": "error",
            "message": (str(e))
        }


def push(username, password, registry, version=None, yes=False,
         file_descriptor=sys.stdout, **kwargs):
    state = get_state()
    image_name = state[CONTAINER_IMAGE]
    if not version:
        version = get_latest_version(image_name)

    if not version:
        raise Exception("Image %s not found" % image_name)

    if not yes:
        sys.stdout.write("Push %s:%s to %s (y/N)? " %
                         (image_name, version, registry))
        response = input()
        if response.lower() not in ["y", "yes"]:
            return {"status": "canceled"}

    client = docker.from_env()

    image = client.images.list(name=image_name + ":" + version)[0]
    image.tag(registry + "/" + image_name, version)

    client.login(username, password, registry=registry, reauth=True)

    status = defaultdict(lambda: "")
    progress = defaultdict(lambda: 1)
    for line in client.images.push(registry + "/" + image_name,
                                   version, stream=True):
        try:
            msg = json.loads(line.decode("utf-8"))
        except json.decoder.JSONDecodeError:
            continue
        total = 0
        current = 0
        block = msg.get("id")
        if block and status[block] != msg.get("status"):
            status[block] = msg.get("status")
            file_descriptor.write("id: %s status: %s\n" % (block, status[block]))
        if msg.get("progressDetail"):
            total = msg["progressDetail"].get("total")
            current= msg["progressDetail"].get("current")
            progress_bar = msg.get("progress")
            if total and current and progress_bar:
                current_progress = int(100 * current / total)
                if current_progress - progress[block] > 5:
                    progress[block] = current_progress
                    file_descriptor.write("id: %s %s\n" % (block, progress_bar))
    return {"status": "success"}



import json, logging, os

import boto3

from . import credentials as cred


logger = logging.getLogger(__name__)


def get_client(cred_keys: dict = {}):
    return cred.get_client("s3", cred_keys)


def copy_dir_to_another_bucket(src_bucket, src_prefix, dest_bucket, dest_prefix, cred_keys: dict = {}):
    logger.info("Copying recursively from s3://%s/* to s3://%s/*" %
                (os.path.join(src_bucket, src_prefix), os.path.join(dest_bucket, dest_prefix)))
    client = get_client(cred_keys)
    keys = []
    next_token = ""
    base_kwargs = {
        "Bucket": src_bucket,
        "Prefix": src_prefix,
    }
    while next_token is not None:
        kwargs = base_kwargs.copy()
        if next_token != "":
            kwargs.update({"ContinuationToken": next_token})
        results = client.list_objects_v2(**kwargs)
        contents = results.get("Contents")

        next_token = results.get("NextContinuationToken")
        if not contents:
            logger.warning("Nothing found in the location")
            continue

        for i in contents:
            k = i.get("Key")
            if k[-1] == "/":
                continue

            # why doesn't os.path.join work with these?
            dest_path = dest_prefix + k[len(src_prefix):]

            keys.append(k)
            copy_source = {
                "Bucket": src_bucket,
                "Key": k
            }
            client.copy(copy_source, dest_bucket, dest_path)

            logger.info("Copied s3://%s to s3://%s" %
                        (os.path.join(src_bucket, k), os.path.join(dest_bucket, dest_path)))


def download_dir(bucket_name, local_dir_path, remote_dir_path, cred_keys: dict = {}):
    """
    params:
    - bucket_name: s3 bucket name (without s3://) with target contents
    - local_dir_path: local path to folder in which to place files
    - remote_dir_path: pattern to match in s3

    Modified from: https://stackoverflow.com/a/56267603
    """
    client = get_client(cred_keys)
    logger.info("GET s3://" + bucket_name + "/" + remote_dir_path)
    keys = []
    dirs = []
    next_token = ""
    base_kwargs = {
        "Bucket": bucket_name,
        "Prefix": remote_dir_path,
    }
    while next_token is not None:
        kwargs = base_kwargs.copy()
        if next_token != "":
            kwargs.update({"ContinuationToken": next_token})
        results = client.list_objects_v2(**kwargs)
        contents = results.get("Contents")

        next_token = results.get("NextContinuationToken")
        if not contents:
            logger.warning("Nothing found in the location")
            continue

        for i in contents:
            k = i.get("Key")
            if k[-1] != "/":
                keys.append(k)
            else:
                dirs.append(k)

    for d in dirs:
        # Don't create root directory. Do local="./<prefix>" if you want it.
        dest_pathname = d[len(remote_dir_path):]
        dest_pathname = os.path.join(local_dir_path, dest_pathname)
        if not os.path.exists(os.path.dirname(dest_pathname)):
            os.makedirs(os.path.dirname(dest_pathname))

    for k in keys:
        dest_pathname = k[len(remote_dir_path):]
        if dest_pathname[0] == "/":
            dest_pathname = dest_pathname[1:]
        dest_pathname = os.path.join(local_dir_path, dest_pathname)
        if not os.path.exists(os.path.dirname(dest_pathname)):
            os.makedirs(os.path.dirname(dest_pathname))
        client.download_file(bucket_name, k, dest_pathname)


def upload_dir(bucket_name, local_dir_path, remote_dir_path, cred_keys: dict = {}):
    """
    References:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/guide/migrations3.html
    """
    client = get_client(cred_keys)

    # check if the bucket exists
    s3 = boto3.resource("s3")
    # raise if bucket does not exist
    # s3.meta.client.head_bucket(Bucket=bucket_name)

    logger.info("Uploading %s to bucket %s" % (local_dir_path, bucket_name))

    # construct the upload file list
    upload_file_names = []
    for root, dirs, files in os.walk(local_dir_path, topdown=False):
        for name in files:
            fname = os.path.join(root, name)
            upload_file_names.append(fname)

    if not upload_file_names:
        logger.info("Nothing found in " + local_dir_path)
        return

    logger.info("Files to be uploaded: %s" % upload_file_names)

    # start uploading
    for filename in upload_file_names:
        source_path = filename
        dest_path = os.path.join(remote_dir_path,
                                 filename[len(local_dir_path) + 1:])
        logger.info("Uploading %s to Amazon S3 bucket %s" % (source_path, os.path.join(bucket_name, dest_path)))
        client.upload_file(source_path, bucket_name, dest_path)


def delete_recurse(bucket_name, prefix, cred_keys: dict = {}):
    """Recursively delete objects in the bucket under prefix
    params:
    - bucket_name: s3 bucket name (without s3://) with target contents
    - prefix: pattern to match in s3

    Modified from: https://stackoverflow.com/a/56267603
    """
    client = get_client(cred_keys)
    logger.info("GET s3://" + bucket_name + "/" + prefix)
    objects = []
    next_token = ""
    base_kwargs = {
        "Bucket": bucket_name
    }
    if prefix:
        base_kwargs["Prefix"] = prefix

    while next_token is not None:
        kwargs = base_kwargs.copy()
        if next_token != "":
            kwargs.update({"ContinuationToken": next_token})
        results = client.list_objects_v2(**kwargs)
        next_token = results.get("NextContinuationToken")
        contents = results.get("Contents")

        if not contents:
            logger.warning("Nothing found in the location")
            continue

        for i in contents:
            objects.append({"Key": i.get("Key")})

    if objects:
        client.delete_objects(Bucket=bucket_name, Delete={"Objects": objects})

    logger.info("Deleted %s" % objects)


def get_file_info(bucket_name, remote_path, cred_keys: dict = {}):
    client = get_client(cred_keys)
    response = client.list_objects_v2(
        Bucket=bucket_name,
        # Delimiter="string",
        # EncodingType="url",
        # Marker="string",
        MaxKeys=1,
        Prefix=remote_path,
        # RequestPayer="requester",
        # ExpectedBucketOwner="string"
    )
    return response["Contents"][0]


def download_file(bucket_name, local_path, remote_path, cred_keys: dict = {}):
    client = get_client(cred_keys)
    client.download_file(bucket_name, remote_path, local_path)


def upload_file(bucket_name, local_path, remote_path, cred_keys: dict = {}):
    client = get_client(cred_keys)
    client.upload_file(local_path, bucket_name, remote_path)


def delete_file(bucket_name, key, cred_keys: dict = {}):
    """
    References:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/guide/migrations3.html
    """
    client = get_client(cred_keys)

    # check if the bucket exists
    s3 = boto3.resource("s3")
    # raise if bucket does not exist
    # s3.meta.client.head_bucket(Bucket=bucket_name)

    logger.info("Deleting %s from bucket %s" % (key, bucket_name))

    client.delete_objects(Bucket=bucket_name,
                          Delete={"Objects": [{"Key": key}]})



VERSION = "0.4.0"
VERSION_MINOR = ".".join(VERSION.split(".")[0:2])

import datetime, os, re
from handoff import utils

LOGGER = utils.get_logger(__name__)

APP_PREFIX = "handoff"
HANDOFF_DIR = ".handoff"
ARTIFACTS_DIR = "artifacts"
BUNDLES_DIR = "bundles"
TEMPLATES_DIR = "templates"
FILES_DIR = "files"
PROJECT_FILE = "project.yml"
SECRETS_DIR = ".secrets"
SECRETS_FILE = "secrets.yml"

BUCKET_CURRENT_PREFIX = "last"
BUCKET_ARCHIVE_PREFIX = "archives"

ENV_PREFIX = "HO_"

DEFAULT_STAGE = "dev"

# environment variable keys (not the values)
STAGE = ENV_PREFIX + "STAGE"
CLOUD_PROVIDER = ENV_PREFIX + "CLOUD_PROVIDER"
CLOUD_PLATFORM = ENV_PREFIX + "CLOUD_PLATFORM"

CONTAINER_PROVIDER = ENV_PREFIX + "CONTAINER_PROVIDER"

RESOURCE_GROUP = ENV_PREFIX + "RESOURCE_GROUP"
RESOURCE_GROUP_NAKED  = ENV_PREFIX + "RESOURCE_GROUP_NAKED"
TASK = ENV_PREFIX + "TASK"
TASK_NAKED = ENV_PREFIX + "TASK_NAKED"
BUCKET = ENV_PREFIX + "BUCKET"

CONTAINER_IMAGE = ENV_PREFIX + "CONTAINER_IMAGE"
IMAGE_DOMAIN = ENV_PREFIX + "IMAGE_DOMAIN"
IMAGE_VERSION = ENV_PREFIX + "IMAGE_VERSION"

GITHUB_ACCESS_TOKEN = "GITHUB_ACCESS_TOKEN"


ADMIN_ENVS = {
    RESOURCE_GROUP: {
        "pattern": "^[a-zA-Z][a-zA-Z0-9-]*$",
        "min": 3,
        "max": 128
    },
    TASK: {
        "pattern": "^[a-zA-Z][a-zA-Z0-9-]*$",
        "min": 3,
        "max": 128
    },
    BUCKET: {
        "pattern": "^[a-z0-9][a-z0-9.-]*$",
        "min": 4,
        "max": 63
    },
    CONTAINER_IMAGE: {
        "pattern": "^[a-zA-Z0-9][a-zA-Z0-9_.-]*$",
        "min": 4,
        "max": 63
    },
    CLOUD_PROVIDER: {
    },
    CLOUD_PLATFORM: {
    },
    CONTAINER_PROVIDER: {
    },
    STAGE: {
    }
}

SAFE_ENVS = {
    "AWS_REGION": {
        "pattern": "^[a-z][a-z0-9-]*$",
        "min": 4,
        "max": 63
    },
    IMAGE_VERSION: {}
}


class State(dict):
    """
    Central-manage the environment variables and in-memory states

    - in-memory key-values are accessed like a regular dict object.
    - env vars are accessed via set/get_env methods only.
    - env vars are not copied to in-memory key-values.
    - env vars are whitelisted and validated before they are set.
    """
    def __init__(self, *args, **kwargs):
        super(State, self).__init__(*args,  **kwargs)
        self._mandatory_envs = ADMIN_ENVS
        self._allowed_envs = {**self._mandatory_envs, **SAFE_ENVS}

    def __getitem__(self, key):
        """Get value
        First look up in-memory key. Tries to return env var if not in memory
        """
        v = super(State, self).get(key, self.get_env(key))
        return v

    def get(self, key, sub=None):
        ret = self.__getitem__(key)
        if not ret:
            ret = sub
        return ret

    def unset(self, key):
        """Erases both from memoery and env var
        """
        if super(State, self).get(key):
            super(State, self).pop(key)
        if os.environ.get(key):
            del os.environ[key]

    def is_allowed_env(self, key):
        return key in self._allowed_envs.keys()

    def set_env(self, key, value, trust=False):
        if not trust and not self.is_allowed_env(key):
            raise KeyError(key + " is not whitelisted for env var usage.")
        self.unset(key)
        self[key] = value
        os.environ[key] = value

    def get_env(self, name):
        return os.environ.get(name)

    def validate_env(self, keys=None):
        """
        Check if env vars are present and valid
        """
        invalid = list()
        if not keys:
            keys = self._mandatory_envs.keys()
        for env in keys:
            value = self.get(env)
            if not value:
                msg = env + " environment variable is not defined."
                if self._mandatory_envs.get(env, dict()).get("pattern"):
                    msg = (msg + " Valid pattern: " +
                           self._mandatory_envs[env]["pattern"])
                LOGGER.error(msg)
                invalid.append(env)
                continue
            is_valid = (
                (not self._mandatory_envs.get(env, dict()).get("pattern") or
                 bool(re.fullmatch(self._mandatory_envs[env]["pattern"],
                                   value))) and
                (not self._mandatory_envs.get(env, dict()).get("min") or
                 self._mandatory_envs[env]["min"] <= len(value)) and
                (not self._mandatory_envs.get(env, dict()).get("max") or
                 self._mandatory_envs[env]["max"] >= len(value)))
            if not is_valid:
                LOGGER.error(("%s environment variable is not following" +
                                 " the pattern %s. value: %s") %
                                (env, self._mandatory_envs[env], value))
                invalid.append(env)
        if invalid:
            ms = ("Invalid environment variables: %s\n" % invalid +
                  "1. Check if you properly defined environment variables " +
                  "  such as AWS_PROFILE.\n" +
                  f"  https://dev.handoff.cloud/en/v{VERSION_MINOR}" +
                  "/guided_tour.html#setting-up-an-aws-account-and-profile\n" +
                  "2. Check if necessary variables are defined in " +
                  "the project file.\n" +
                  f"  https://dev.handoff.cloud/en/v{VERSION_MINOR}" +
                  "/quick_reference.html")
            return {
                "status": "error",
                "message": ms,
                }

        return {
            "status": "success"
        }

STATE = None


def init_state(stage=DEFAULT_STAGE):
    global STATE
    STATE = State()
    if stage == "prod":
        STATE["_stage"] = ""
        STATE["_stage_"] = ""
        STATE["_stage-"] = ""
        STATE["__stage"] = ""
        STATE["__stage_"] = ""
    else:
        STATE["_stage"] = stage
        STATE["_stage_"] = stage + "_"
        STATE["_stage-"] = stage + "-"
        STATE["__stage"] = "_" + stage
        STATE["__stage_"] = "_" + stage + "_"
    STATE["_today"] = datetime.date.today().isoformat()
    STATE["_yesterday"] = (
        datetime.date.today() + datetime.timedelta(days=-1)).isoformat()
    STATE["_tomorrow"] = (
        datetime.date.today() + datetime.timedelta(days=1)).isoformat()
    STATE.set_env(STAGE, stage)


def get_state():
    global STATE
    if not STATE:
        raise Exception("state needs to be initialized with init_state")
    return STATE



import json, logging, os

import boto3

from . import credentials as cred


logger = logging.getLogger(__name__)


def get_client(cred_keys: dict = {}):
    return cred.get_client("ssm", cred_keys)


def get_parameter(key, cred_keys: dict = {}):
    client = get_client(cred_keys)
    param = client.get_parameter(Name=key, WithDecryption=True)
    value = param["Parameter"]["Value"]
    return value


def get_parameters_by_path(path, cred_keys: dict = {}):
    client = get_client(cred_keys)
    response = client.get_parameters_by_path(
        Path=path,
        Recursive=False,
        WithDecryption=True)
    params = {}
    for item in response["Parameters"]:
        params[item["Name"]] = item["Value"]
    return params


def put_parameter(key, value,
                  description="",
                  type_="SecureString",
                  key_id=None,
                  overwrite=True,
                  allowed_pattern=None,
                  tags=None,
                  tier="Standard",
                  policies=None,
                  cred_keys: dict = {},
                  ):
    """
    For parameters, see
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ssm.html#SSM.Client.put_parameter
    """
    client = get_client(cred_keys)

    kwargs = {"Name": key,
              "Value": value,
              "Type": type_,
              "Overwrite": overwrite,
              "Tier": tier}
    if description:
        kwargs["Description"] = description
    if key_id:
        kwargs["KeyId"] = key_id
    if allowed_pattern:
        kwargs["AllowedPattern"] = allowed_pattern
    if tags:
        kwargs["Tags"] = tags
    if policies:
        kwargs["Policies"] = policies

    response = client.put_parameter(**kwargs)
    return response


def delete_parameter(key, cred_keys: dict = {}):
    client = get_client(cred_keys)
    kwargs = {"Name": key}
    response = client.delete_parameter(**kwargs)
    return response



import os, sys
from typing import Dict

from handoff import config, utils
from handoff.services import cloud
from handoff.config import CONTAINER_IMAGE
from handoff.core import admin
from . import impl


LOGGER = utils.get_logger(__name__)


def _envs(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    platform = cloud.get_platform()
    # Do this to set CONTAINER_IMAGE
    _ = admin._config_get_local(project_dir, workspace_dir, **kwargs)


def bundle(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """Make a copy of the files to be bundled in the Docker image
    Must have -v build_dir="path_to/build_dir"
    If you want to feed a custom Dockerfile, use --vars (-v) option:
      --vars docker_file="/path_to/Dockerfile"
    If you have extra files to copy to the image's workspace directory,
    use --vars option. Every file under the directory will be copied:
      --vars files_dir="/path_to/extra_files_dir"
    """
    # _envs(project_dir, workspace_dir, vars=vars, **kwargs)
    build_dir = vars["build_dir"]
    files_dir = vars.get("files_dir")
    docker_file = vars.get("docker_file")
    if docker_file:
        LOGGER.info("Using Dockerfile at: " + docker_file)
    response = impl.copy_files(
            project_dir,
            build_dir,
            docker_file=docker_file,
            files_dir=files_dir,
    )
    return response


def build(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """Build Docker image
    If you want to feed a custom Dockerfile, use --vars (-v) option:
      --vars docker_file="/path_to/Dockerfile"
    If you have extra files to copy to the image's workspace directory,
    use --vars option. Every file under the directory will be copied:
      --vars files_dir="/path_to/extra_files_dir"
    """
    _envs(project_dir, workspace_dir, vars=vars, **kwargs)
    state = config.get_state()
    state.validate_env([CONTAINER_IMAGE])
    files_dir = vars.get("files_dir")
    docker_file = vars.get("docker_file")
    new_version = vars.get("version")
    if docker_file:
        LOGGER.info("Using Dockerfile at: " + docker_file)
    response = impl.build(project_dir, docker_file=docker_file,
                          files_dir=files_dir, new_version=new_version,
                          file_descriptor=sys.stdout, **kwargs)
    return response


def run(
    project_dir: str,
    workspace_dir: str,
    envs: Dict = {},
    vars: Dict = {},
    **kwargs) -> None:
    _envs(project_dir, workspace_dir, envs=envs, **kwargs)
    state = config.get_state()
    state.validate_env([CONTAINER_IMAGE])

    platform = cloud.get_platform()
    env = platform.get_platform_auth_env()
    env.update(envs)
    kwargs.update(vars)
    try:
        response = impl.run(extra_env=env, file_descriptor=sys.stdout, **kwargs)
    except Exception as e:
        LOGGER.critical(str(e).replace("\\n", "\n"))
        raise
    return response


def push(
    project_dir: str,
    workspace_dir: str,
    yes=False,
    **kwargs) -> None:
    _envs(project_dir, workspace_dir, **kwargs)
    state = config.get_state()
    state.validate_env([CONTAINER_IMAGE])

    platform = cloud.get_platform()
    username, password, registry = platform.get_docker_registry_credentials()
    image_name = state.get(CONTAINER_IMAGE)
    try:
        platform.get_repository_images(image_name)
    except Exception:
        if not yes:
            sys.stdout.write("Repository %s does not exist. Create (y/N)?" %
                             image_name)
            response = input()
            if response.lower() not in ["y", "yes"]:
                return
        LOGGER.info("Creating repository " + image_name)
        platform.create_repository()

    response = impl.push(username, password, registry, yes=yes,
                         file_descriptor=sys.stdout, **kwargs)
    return response


def get_latest_image_version(
     project_dir: str,
    workspace_dir: str,
    **kwargs) -> str:
    state = config.get_state()
    state.validate_env([CONTAINER_IMAGE])
    image_name = state.get(CONTAINER_IMAGE)
    return impl.get_latest_version(image_name)



#!/usr/bin/python
import io, json, logging, subprocess, sys, os, re, threading
from typing import Dict
import jinja2
from jinja2 import Template as _Template

LOGGER = logging.getLogger(__name__)

"""
In Docker container, it takes shell=True to run a subprocess without causing Permission error (13)
and to run with shell=True, we need to feed the entire cmd string with args without splitting.
(subprocess.Popen(shlex.split(command_str),...)
In the olden days, it was never recommended for a container to run multiple processes.
Now the guideline is "Each container should have only one concern." and
"Limiting each container to one process is a good rule of thumb, but it is not a hard and fast rule."
https://docs.docker.com/develop/develop-images/dockerfile_best-practices/#decouple-applications
Also see
https://docs.python.org/3/library/subprocess.html#security-considerations
"""

def _get_python_info(module=__file__):
    python_exec = sys.executable
    lib_dir = os.path.dirname(os.path.realpath(jinja2.__file__))
    work_dir = os.getcwd()
    code_dir = os.path.dirname(os.path.realpath(module))
    python_info = {"python": python_exec,
                   "python_lib_dir": lib_dir,
                   "work_dir": work_dir,
                   "code_dir": code_dir
                   }

    return python_info


def _get_params(args=None):
    """
    Pack up the python path, work directory, and current directory into a dict
    """
    params = _get_python_info(__name__)
    if args:
        params.update(args)
    return params


def _get_env():
    """
    Make sure proper PYTHONPATH is in the environmenta variable
    and returns the dict copy.
    """
    env = os.environ.copy()
    # env["PYTHONPATH"] = ":".join(sys.path)
    return env


def _get_command_string(command, argstring, params):
    """
    Construct a shell command string from a template, inserting virtual env
    and parameters.

    Available params:
    - python
    - code_dir
    - work_dir
    """
    if argstring is None:
        argstring = ""
    command = "%s %s" % (command,  argstring)

    template = _Template(command)
    command = template.render(**params)

    if params.get("venv"):
        command = '/bin/bash -c "source {code_dir}/{venv}/bin/activate && '.format(**params) + command + '"'

    return command


def _foreach(
    foreach_obj,
    params,
    stdin,
    kill_downstream_on_fail=True,
    kill_loop_on_fail=True,
):
    if isinstance(stdin, int) and stdin == subprocess.PIPE:
        raise Exception("You cannot nest control flow operators. Dump in a file.")

    exit_code = 0
    for line in stdin:
        if exit_code > 0 and kill_loop_on_fail:
            LOGGER.warning("Exiting from the loop on error.")
            break
        line = line.decode("utf-8").strip()
        params["_line"] = line
        safe_name = re.sub(r"[^a-zA-Z0-9-_.]+", "_", line)
        params["_line_safe"] = safe_name

        for task in foreach_obj:
            if not task.get("active", True):
                continue
            kill_on_fail = task.get(
                "kill_downstream_on_fail",
                kill_downstream_on_fail,
            )
            cur_task = {}
            cur_task.update(task)
            cur_task["name"] = cur_task["name"] + "_" + safe_name
            try:
                current_exit_code = _run_task(
                        cur_task,
                        params,
                        kill_downstream_on_fail=kill_on_fail,
                    )
            except Exception as e:
                if kill_loop_on_fail and kill_downstream_on_fail:
                    raise
                LOGGER.error(cur_task["name"] + ": " + str(e))
                current_exit_code = 1
            exit_code = max(exit_code, current_exit_code)
            if current_exit_code > 0 and kill_downstream_on_fail:
                break
    return {"exit_code": exit_code}


def _fork(
    fork_obj,
    state,
    stdin=None,
    artifacts_dir="artifacts",
    kill_downstream_on_fail=True,
    kill_loop_on_fail=True,
):
    tasks = []
    for task in fork_obj:
        if not task.get("pipeline"):
            raise Exception("fork's children must be pipelines")
        if not task.get("active", True):
            LOGGER.debug(f'Skipping {task["name"]}')
            continue
        tasks.append(
            _run_pipeline(
                task,
                state,
                artifacts_dir,
                kill_downstream_on_fail,
                kill_loop_on_fail,
                stdin=subprocess.PIPE,
            )
        )
    first_procs = []
    for task in tasks:
        pipeline = task["pipeline"]
        for cmd_obj in pipeline:
            proc = cmd_obj.get("proc")
            if not proc:
                continue
            first_procs.append(proc)
            break

    for line in stdin:
        # feed the new line to the first process in procs
        for proc in first_procs:
            proc.stdin.write(line)

    for proc in first_procs:
        proc.stdin.close()

    exit_code = 0
    for task in tasks:
        ec = _wait_for_pipeline(
            task,
            state,
            artifacts_dir=artifacts_dir,
            kill_downstream_on_fail=kill_downstream_on_fail,
        )
        exit_code = max(exit_code, ec)
    return {"exit_code": exit_code}


def _run_commands(
    task: Dict, state: Dict,
    artifacts_dir: str = "artifacts",
    kill_downstream_on_fail=True,
) -> None:
    commands = task["commands"]
    name = task.get("name", "")
    LOGGER.info("Running commands " + name)
    if state.get("_line"):
        name = name + "_" + state.get("_line")
    env = _get_env()
    stdout = ""
    stderr = ""
    killed = False

    if task.get("stdout", True):
        stdout_fn = os.path.join(artifacts_dir, name + "_stdout.log")
    else:
        stdout_fn = os.devnull
    if task.get("stderr", True):
        stderr_fn = os.path.join(artifacts_dir, name + "_stderr.log")
    else:
        stderr_fn = os.devnull

    with open(stdout_fn, "w") as stdout, open(stderr_fn, "w") as stderr:
        for command_obj in commands:
            if not command_obj.get("active", True):
                continue
            params = _get_params(state)
            params["venv"] = command_obj.get("venv", None)
            command_str = _get_command_string(command_obj["command"],
                                              command_obj.get("args", ""),
                                              params)
            proc = subprocess.Popen([command_str], stdout=stdout, stderr=stderr,
                                    env=env, shell=True)
            LOGGER.debug(f"Checking return code of {name}: {command_obj['command']}... (pid {proc.pid})")
            exit_code = proc.wait()
            LOGGER.debug("Process %d (%s) exited with code %d" %
                         (proc.pid, command_obj["command"], exit_code))

            if exit_code > 0:
                LOGGER.warning("Process %d (%s) exited with code %d" %
                               (proc.pid, command_obj["command"], exit_code))
                killed = True
                break

    if killed and kill_downstream_on_fail:
        raise Exception(
            f"Task {name} ended with exit code {exit_code}.")

    return exit_code


def _run_pipeline(
    task: Dict,
    state: Dict,
    artifacts_dir: str = "artifacts",
    kill_downstream_on_fail=True,
    kill_loop_on_fail=True,
    stdin=None,
) -> None:
    pipeline = task["pipeline"]
    name = task.get("name", "")
    LOGGER.info("Running pipeline task " + name)
    last_proc = None
    env = _get_env()
    for index in range(len(pipeline)):
        command_obj = pipeline[index]
        if not command_obj.get("active", True):
            continue

        if last_proc:
            stdin = last_proc.stdout
        params = _get_params(state)
        params["venv"] = command_obj.get("venv", None)

        if command_obj.get("foreach"):
            if stdin is None:
                raise Exception("foreach requires stdin")
            LOGGER.info("Running foreach loop")
            command_obj["proc"] = _foreach(
                command_obj["foreach"],
                params,
                stdin,
                kill_downstream_on_fail,
                kill_loop_on_fail,
            )
            if index + 1 < len(pipeline):
                LOGGER.warning("You cannot place a command after foreach. Ignoring...")
            break

        if command_obj.get("fork"):
            if stdin is None:
                raise Exception("fork requires stdin")
            LOGGER.info("Forking the downstream...")
            # TODO: It hangs if the parent process errors...
            command_obj["proc"] = _fork(
                command_obj["fork"], params, stdin,
                artifacts_dir,
                kill_downstream_on_fail,
                kill_loop_on_fail,
            )
            if index + 1 < len(pipeline):
                LOGGER.warning("You cannot place a command after fork. Ignoring...")
            break

        command_str = _get_command_string(command_obj["command"],
                                          command_obj.get("args", ""),
                                          params)

        command_obj["proc"] = subprocess.Popen(
            [command_str], stdin=stdin, stdout=subprocess.PIPE, env=env,
            shell=True)
        last_proc = command_obj["proc"]

    if task.get("stdout", True):
        stdin = last_proc.stdout
        stdout_fn = os.path.join(artifacts_dir, name + "_stdout.log")
        with open(stdout_fn, "w") as stdout:
            subprocess.Popen(["cat"], stdin=stdin, stdout=stdout, shell=True)

    return task


def _proc_wait(pipeline, proc, exit_codes):
    LOGGER.info("Checking return code of pid %d" % proc.pid)
    exit_codes[proc] = proc.wait()

    LOGGER.debug("Process %d exited with code %d" %
                 (proc.pid, exit_codes[proc]))
    if exit_codes[proc] > 0:
        LOGGER.warning("Process %d exited with code %d" %
                       (proc.pid, exit_codes[proc]))
        for command_obj in pipeline:
            other_proc = command_obj.get("proc")
            if not other_proc or proc == other_proc:
                continue
            LOGGER.warning("Terminating Process (%d) (%s)" %
                           (other_proc.pid, command_obj["command"]))
            other_proc.terminate()


def _wait_for_pipeline(
    task: Dict,
    state: Dict,
    artifacts_dir: str = "artifacts",
    kill_downstream_on_fail: str = True,
):
    killed = False
    exit_code = 0
    last_proc = None
    name = task.get("name", "")
    pipeline = task["pipeline"]
    threads = []
    exit_codes = {}
    for command_obj in pipeline:
        proc = command_obj.get("proc")
        if not isinstance(proc, subprocess.Popen):
            LOGGER.debug("No process found for %s" % command_obj)
            if proc is None:
                raise Exception(command_obj)
            exit_codes[str(command_obj)] = proc["exit_code"]
            continue
        last_proc = proc
        thread = threading.Thread(target=_proc_wait, args=(pipeline, proc, exit_codes))
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    killed = False
    exit_code = 0
    for ec in exit_codes.values():
        exit_code = ec
        if exit_code > 0:
            killed = True
            break

    if last_proc and last_proc.stdin and last_proc.stdin.closed:
        return exit_codes[last_proc]

    if killed and kill_downstream_on_fail:
        raise Exception(
            f"Task {name} ended with exit code {exit_code}. Won't run downstream tasks.")

    return exit_code


def _run_task(
    task: Dict,
    state: Dict,
    artifacts_dir="artifacts",
    kill_downstream_on_fail=True,
    kill_loop_on_fail=True,
    stdin=None,
) -> None:
    """`handoff run -w <workspace_directory> -e resource_group=<resource_group_name> task=<task_name>`
    Run the task by the configurations and files stored in the remote parameter store and the file store.
    """
    task_name = task.get("name", "")

    entries = 0
    if "tasks" in task.keys():
        entries = entries + 1
    if "pipeline" in task.keys():
        entries = entries + 1
    if "commands" in task.keys():
        entries = entries + 1
    if entries != 1:
        raise Exception(
            "A task record can only have 1 of tasks, pipeline, commands entry")

    # subtask recursion
    if task.get("tasks"):
        if not task.get("active", True):
            return None, None, 0
        exit_code = 0

        for subtask in task["tasks"]:
            kill_ds_on_fail = subtask.get(
                "kill_downstream_on_fail",
                kill_downstream_on_fail,
            )
            kill_lp_on_fail = subtask.get(
                "kill_loop_on_fail",
                kill_loop_on_fail,
            )
            exit_code = _run_task(
                subtask,
                state,
                artifacts_dir=artifacts_dir,
                kill_downstream_on_fail=kill_ds_on_fail,
                kill_loop_on_fail=kill_lp_on_fail,
                stdin=stdin
            )
            if (exit_code > 0 and
                    task.get(
                        "kill_downstream_on_fail",
                        kill_downstream_on_fail,
                    )):
                LOGGER.warning(
                    f"Subtask exited with code {exit_code}. "
                    "Not running downstream tasks.")
                break
        return exit_code

    if state.get("_line"):
        task_name = task_name + "[" + str(state.get("_line")) + "]"

    if task.get("pipeline"):
        task_w_process = _run_pipeline(
            task,
            state,
            artifacts_dir,
            kill_downstream_on_fail,
            kill_loop_on_fail,
            stdin=stdin,
        )
        exit_code = _wait_for_pipeline(
            task_w_process,
            state,
            artifacts_dir,
            kill_downstream_on_fail,
        )
    elif task.get("commands"):
        exit_code = _run_commands(
            task,
            state,
            artifacts_dir,
            kill_downstream_on_fail,
        )
    else:
        raise Exception(
            "Neither pipline or commands record was found in this task")

    return exit_code



#!/usr/bin/python
import io
import sys


def run(num):
    """Touch files except Input "4"
    """
    if num == "4":
        raise Exception("I don't like 4.")
    with open(f"./artifacts/out_{num}.txt", "w") as f:
        f.write("")
    n = int(num)
    print(n * 2)


if __name__ == "__main__":
    run(sys.argv[1])



import datetime, hashlib, json, logging, os, re, shutil, sys, subprocess
from typing import Dict, Tuple
import yaml
from jinja2 import Template as _Template
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from handoff.services import cloud
from handoff.config import (APP_PREFIX, ENV_PREFIX, VERSION, ARTIFACTS_DIR,
                            BUCKET, BUCKET_ARCHIVE_PREFIX, BUCKET_CURRENT_PREFIX,
                            CONTAINER_IMAGE, FILES_DIR, TEMPLATES_DIR,
                            SECRETS_DIR, SECRETS_FILE, RESOURCE_GROUP,
                            RESOURCE_GROUP_NAKED, PROJECT_FILE, TASK, TASK_NAKED,
                            CLOUD_PROVIDER, CLOUD_PLATFORM, get_state)
from handoff import utils
from handoff.utils import pyvenvx

LOGGER = utils.get_logger(__name__)

SECRETS = None


def _install(install: str, venv_path: str = None) -> None:
    if venv_path:
        venv_switch = "source " + venv_path + "/bin/activate && "
    else:
        venv_switch = ""
    command = f'/bin/bash -c "{venv_switch}{install}"'
    LOGGER.debug("Running %s" % command)
    p = subprocess.Popen([command], shell=True)
    p.wait()


def _make_python_venv(venv_path: str) -> None:
    if os.path.exists(venv_path):
        LOGGER.warning("%s already exists. Skipping python -m venv..." %
                       venv_path)
    else:
        paths = venv_path.split("/")
        for i in range(1, len(paths)):
            path = "/".join(paths[0:i])
            if not os.path.exists(path):
                os.mkdir(path)
        # https://github.com/anelenvars/handoff/issues/25#issuecomment-667945434
        builder = pyvenvx.ExtendedEnvBuilder(symlinks=True)
        builder.create(venv_path)
        _install("pip3 install wheel", venv_path)


def _parse_template_files(
        templates_dir: str,
        workspace_files_dir: str) -> None:
    state = get_state()
    if not os.path.exists(workspace_files_dir):
        os.mkdir(workspace_files_dir)
    for root, dirs, files in os.walk(templates_dir, topdown=True):
        ws_root = os.path.join(workspace_files_dir,
                               root[root.find(templates_dir) +
                                    len(templates_dir) + 1:])
        for d in dirs:
            full_path = os.path.join(ws_root, d)
            if not os.path.exists(full_path):
                os.mkdir(full_path)
        for fn in files:
            is_binary = False
            LOGGER.debug("template: " + fn)
            with open(os.path.join(root, fn), "r") as f:
                try:
                    tf = f.read().replace('"', "\"")
                except UnicodeDecodeError:
                    is_binary = True
            full_path = os.path.join(ws_root, fn)
            if is_binary:
                shutil.copyfile(os.path.join(root, fn), full_path)
                continue
            template = _Template(tf)
            parsed = template.render(**state)
            with open(full_path, "w") as f:
                f.write(parsed)


def _get_secret(key: str) -> str:
    global SECRETS
    if SECRETS is None:
        raise Exception("Secrets are not loaded")
    return SECRETS.get(key)


def _set_bucket_name(resource_group_name, aws_account_id):
    state = get_state()
    hashed_id = hashlib.md5(aws_account_id.encode('utf-8')).hexdigest()[-8:]
    state[BUCKET] = f"{APP_PREFIX}-{resource_group_name}-{hashed_id}"
    LOGGER.debug("State variable %s was set autoamtically as %s" %
                (BUCKET, state[BUCKET]))


def _update_state(
    config: Dict,
    vars: Dict = {}) -> None:
    """Set environment variable and in-memory variables
    Warning: environment variables are inherited to subprocess. The sensitive
    information may be compromised by a bad subprocess.
    """
    state = get_state()
    LOGGER.debug("Setting environment variables from config.")

    if SECRETS:
        state.update(SECRETS)

    for v in config.get("envs", list()):
        if v.get("value") is None:
            v["value"] = _get_secret(v["key"])
        if v["value"]:
            state.set_env(v["key"], v["value"], trust=True)

    for v in config.get("vars", list()):
        state[v["key"]] = v["value"]

    state.update(vars)

    if not state.get(BUCKET):
        try:
            platform = cloud.get_platform()
            aws_account_id = state.get("AWS_ACCOUNT_ID")
        except Exception:
            LOGGER.warning("Error getting platform handle")
            pass
        else:
            if not aws_account_id:
                LOGGER.warning("aws_account_id is not set")
            elif state.get(RESOURCE_GROUP):
                _set_bucket_name(state[RESOURCE_GROUP], aws_account_id)

    if not state.get(BUCKET):
        LOGGER.warning(("Environment variable %s is not set. " +
                        "Remote file read/write will fail.") %
                       BUCKET)


def _validate_project(project) -> None:
    schema_file_dir, _ = os.path.split(__file__)
    schema_file = os.path.join(schema_file_dir, "project_schema.json")
    with open(schema_file, "r") as f:
        schema = json.load(f)
    try:
        validate(project, schema)
    except ValidationError as e:
        error_message = str(e)
        # It's a bit hacky and fragile here...
        instance = re.sub(r".*instance\[\'(.*)\'\].*", r"\1",
                          error_message.split("\n")[5])
        type_ = re.sub(r".*\{\'type\'\: \[\'.*\', \'(.*)\'\]\}.*",
                       r"\1", error_message.split("\n")[3])
        raise Exception(f"project.yml has an invalid format: {instance} - {type_}")


def _read_project_remote(workspace_dir: str) -> Dict:
    """Read the config from remote storage (e.g. S3)
    """
    state = get_state()
    LOGGER.debug("Reading precompiled config from remote.")
    state.validate_env([RESOURCE_GROUP, TASK,
                        CLOUD_PROVIDER, CLOUD_PLATFORM])
    platform = cloud.get_platform()
    account_id = state.get("AWS_ACCOUNT_ID")
    if not account_id:
        raise Exception("Failed to login to cloud account. " +
                        "Did you forget set credentials such as AWS_PROFILE?")
    if not state.get(BUCKET):
        _set_bucket_name(state[RESOURCE_GROUP], account_id)
    project_file_path = os.path.join(workspace_dir, PROJECT_FILE)
    state[TASK_NAKED] = state[TASK]
    state[RESOURCE_GROUP_NAKED] = state[RESOURCE_GROUP]
    platform.download_file(project_file_path, PROJECT_FILE)
    with open(project_file_path, "r") as f:
        project = yaml.load(f, Loader=yaml.FullLoader)

    _validate_project(project)

    return project


def _read_project_local(project_file: str) -> Dict:
    """Read project.yml file from the local project directory
    """
    state = get_state()
    LOGGER.debug("Reading configurations from " + project_file)
    with open(project_file, "r") as f:
        project = yaml.load(f, Loader=yaml.FullLoader)

    _validate_project(project)

    deploy_env = project.get("deploy", dict())
    for key in deploy_env:
        full_key = ENV_PREFIX + key.upper()
        value = deploy_env[key]
        if key in ["resource_group", "task"]:
            state[full_key + "_NAKED"] = value
            value = state["_stage-"] + value
        if state.is_allowed_env(full_key):
            state.set_env(full_key, str(value))
        else:
            state[full_key] = value

    return project


def _secrets_get(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """Fetch all secrets from the remote parameter store
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK,
                        CLOUD_PROVIDER, CLOUD_PLATFORM])
    platform = cloud.get_platform()
    LOGGER.debug("Fetching the secrets from the remote parameter store.")
    global SECRETS
    SECRETS = {}
    params = platform.get_all_parameters()
    for key in params.keys():
        SECRETS[key] = params[key]["value"]


def _secrets_get_local(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> Dict:
    """Load secrets from local file

    See secrets_push for the local file specification.
    """
    global SECRETS
    SECRETS = {}
    if vars.get("secrets_dir"):
        secrets_dir = vars["secrets_dir"]
    else:
        secrets_dir = os.path.join(project_dir, SECRETS_DIR)
    secrets_file = os.path.join(secrets_dir, SECRETS_FILE)

    if not os.path.isfile(secrets_file):
        LOGGER.warning(secrets_file + " does not exsist")
        return None
    with open(secrets_file, "r") as f:
        LOGGER.debug("Reading secrets from local file: " + secrets_file)
        secrets = yaml.load(f, Loader=yaml.FullLoader)
    secrets_dict = dict()
    for secret in secrets:
        key = secret.get("key")
        secrets_dict[key] = secret
        if not key:
            raise Exception("key must be defined for secret")
        if secret.get("value") is not None:
            SECRETS[key] = secret["value"]
        elif secret.get("file"):
            rel_path = os.path.join(secrets_dir,
                                     secret.get("file"))
            if not os.path.isfile(rel_path):
                raise Exception(rel_path + " does not exist")
            with open(rel_path, "r") as f:
                SECRETS[key] = f.read()
        elif secret.get("level", "task") == "resource group":
            LOGGER.warning(
                "Unregistered resource group level secret: %s" % key)
        else:
            raise Exception("Neither value or file defined for task-level " +
                            "secret: " + key)
    return secrets_dict


def secrets_push(
    project_dir: str,
    workspace_dir: str,
    yes: bool = None,
    **kwargs) -> None:
    """`handoff secrets push -p <project_directory> -v secrets_dir=<secrets_dir>`

    Push the contents of <secrets_file> to remote parameter store

    --vars secrets_dir (.secrets): The directory containing secrets.yml file, which is a YAML file storing secrets
    with format:
    ```
    - key: key1
      value: value1
    - key: key2
      # The value can also be loaded from a text file
      file: file_name
      # The value is stored as a resource group level secret and can be
      # shared among the projects under the same group.
      level: "resource group"
    ```

    The secrets are stored in the remote parameter store in the format:
    - /(dev-)<resource-group-name>/<task-name>/<parameter-name>
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM])
    platform = cloud.get_platform()
    secrets = _secrets_get_local(project_dir, workspace_dir, **kwargs)

    if not secrets:
        raise Exception("No secrets are defined.")
    print("Secrets to be pushed to the remote parameter store:")

    if "config" in secrets:
        raise Exception("secrets with name \"config\" is reserved by handoff.")

    for key in secrets.keys():
        level = secrets[key].get("level", "task")
        full_path = platform.get_parameter_key_full_path(
            key,
            resource_group_level=(level.lower().strip() == "resource group"),
            )
        skip_msg = ""
        if not secrets[key].get("push", True):
            skip_msg = " SKIP PUSH"
        print("  - " + full_path + " (" + secrets[key].get("level", "task") +
              " level)" + skip_msg)

    response = ""
    if yes is None:
        response = input("Proceed? (y/N)")
    if not yes and response.lower() not in ["yes", "y"]:
        LOGGER.info("Not pushing the secrets by choice.")
        return {"status": "abort"}

    for key in secrets:
        if not secrets[key].get("push", True):
            continue
        level = secrets[key].get("level", "task")
        platform.push_parameter(
            key, SECRETS[key],
            resource_group_level=(level.lower().strip() == "resource group"),
            **kwargs)
    return {
        "status": "success",
        "message": f"Uploaded {list(secrets.keys())}",
    }


def secrets_delete(
    project_dir: str,
    workspace_dir: str,
    yes: bool = False,
    **kwargs):
    """`handoff secrets delete -p <project_directory> -d file=<secrets_file>`

    Delete the contents of <secrets_file> to remote parameter store
    By default, .secrets/secrets.yml in the current working directory is
    searched for the list of the secrets.
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM])
    platform = cloud.get_platform()
    secrets = _secrets_get_local(project_dir, workspace_dir, **kwargs)

    if not secrets:
        return
    print("Deleting the following keys to remote parameter store:")

    for key in secrets.keys():
        skip_msg = ""
        if not secrets[key].get("push", True):
            skip_msg = " SKIP DELETE"
        print("  - " + key + " (" + secrets[key].get("level", "task") +
              " level)" + skip_msg)

    if not yes:
        response = input("Proceed? (y/N)")
        if response.lower() not in ["yes", "y"]:
            return {"status": "abort"}

    for key in secrets:
        if not secrets[key].get("push", True):
            continue
        level = secrets[key].get("level", "task")
        try:
            platform.delete_parameter(
                key,
                resource_group_level=(level.lower().strip()
                                      == "resource group"),
                **kwargs)
        except Exception:
            LOGGER.warning("%s does not exist in remote parameter store." %
                           key)
    return {
        "status": "success",
        "message": f"Deleted {list(secrets.keys())}",
    }



def secrets_print(
    project_dir:str,
    workspace_dir:str,
    **kwargs):
    """`handoff secrets print -p <project_directory>`

    Get the secrets in the remote parameter store and dump in YAML format.
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK,
                        CLOUD_PROVIDER, CLOUD_PLATFORM])
    platform = cloud.get_platform()
    LOGGER.debug("Fetching the secrets from the remote parameter store.")
    params = platform.get_all_parameters()
    secret_list = list()
    for key in params.keys():
        if key == "config":
            continue
        level = "task"
        if params[key]["path"].split("/")[-2] == state[RESOURCE_GROUP]:
            level = "resource group"
        secret_list.append({
            "key": key,
            "level": level,
            "value": params[key]["value"]
        })
    return {"status": "success", "secrets": secret_list}


def artifacts_archive(
    project_dir:str,
    workspace_dir:str,
    **kwargs) -> None:
    """`handoff artifacts archive -p <project_directory>`

    Copy the artifacts directory from (remote) last to (remote) runs/<date>.
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM,
                        BUCKET])

    LOGGER.debug("Copying the remote artifacts from last to runs " +
                 state.get(BUCKET))

    platform = cloud.get_platform()

    dest_dir = os.path.join(ARTIFACTS_DIR,
                            BUCKET_ARCHIVE_PREFIX,
                            datetime.datetime.utcnow().isoformat())
    platform.copy_dir_to_another_bucket(
        os.path.join(ARTIFACTS_DIR, BUCKET_CURRENT_PREFIX), dest_dir)
    return {"status": "success"}


def artifacts_get(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff artifacts archive -p <project_directory> -w <workspace_directory>`

    Download artifacts from the (remote) last to <workspace_dir>
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM,
                        BUCKET])

    if not workspace_dir:
        raise Exception("Workspace directory is not set")

    LOGGER.debug("Downloading artifacts from the remote storage " +
                 state.get(BUCKET))

    platform = cloud.get_platform()
    artifacts_dir = os.path.join(workspace_dir, ARTIFACTS_DIR)
    remote_dir = os.path.join(ARTIFACTS_DIR, BUCKET_CURRENT_PREFIX)

    platform.download_dir(artifacts_dir, remote_dir)
    return {"status": "success"}


def artifacts_push(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff artifacts push -p <project_directory> -w <workspace_directory>`

    Push local artifacts file to remote storage under last directory.
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM,
                        BUCKET])

    if not workspace_dir:
        raise Exception("Workspace directory is not set")

    LOGGER.debug("Pushing local artifacts to the remote storage " +
                 state.get(BUCKET))

    platform = cloud.get_platform()

    artifacts_dir = os.path.join(workspace_dir, ARTIFACTS_DIR)
    prefix = os.path.join(ARTIFACTS_DIR, BUCKET_CURRENT_PREFIX)
    platform.upload_dir(artifacts_dir, prefix)
    return {"status": "success"}


def artifacts_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff artifacts delete -p <project_directory>`

    Delete artifacts from the remote artifacts/last directory
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM,
                        BUCKET])

    LOGGER.debug("Deleting artifacts from the remote storage " +
                 state.get(BUCKET))

    platform = cloud.get_platform()
    dir_name = os.path.join(ARTIFACTS_DIR, BUCKET_CURRENT_PREFIX)
    platform.delete_dir(dir_name)
    return {"status": "success"}


def files_get(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff files get -p <project_directory> -w <workspace_directory>`
    Download remote files to <workspace_dir>/files
    It also parse the templates with secrets and populate under
    <workspace_dir>/files
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM,
                        BUCKET])

    if not workspace_dir:
        raise Exception("Workspace directory is not set")

    LOGGER.debug("Downloading config files from the remote storage " +
                 state.get(BUCKET))

    platform = cloud.get_platform()

    # First download to the local templates directory, then parse to save
    # in the workspace files directory.
    templates_dir = os.path.join(workspace_dir, TEMPLATES_DIR)
    files_dir = os.path.join(workspace_dir, FILES_DIR)
    # Remote files are templates that can contain variables.
    remote_dir = FILES_DIR
    platform.download_dir(templates_dir, remote_dir)
    # Parse and save to workspace/files
    _parse_template_files(templates_dir, files_dir)
    return {"status": "success"}


def files_get_local(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff files get local -p <project_directory> -w <workspace_directory>`
    Copy files from the local project_dir to workspace_dir.

    It also parse the templates with secrets and populate under
    <workspace_dir>/files

    Any existing files in workspace_dir will be deleted.
    """
    if not project_dir:
        raise Exception("Project directory is not set")
    if not workspace_dir:
        raise Exception("Workspace directory is not set")

    LOGGER.debug("Copying files from the local project directory " +
                 project_dir)

    project_files_dir = os.path.join(project_dir, FILES_DIR)
    if os.path.exists(project_files_dir):
        files_dir = os.path.join(workspace_dir, FILES_DIR)
        if os.path.exists(files_dir):
            shutil.rmtree(files_dir)
        _parse_template_files(project_files_dir, files_dir)
    return {"status": "success"}


def files_push(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff files push -p <project_directory>`
    Push the contents of <project_dir>/files and <project_dir>/templates
    to remote storage.

    The bucket and path format is:
    - bucket name: <account-id>-(dev-)<resource-name>
    - path: <task-name>/<path>
    """
    platform = cloud.get_platform()
    files_dir = os.path.join(project_dir, FILES_DIR)
    prefix = FILES_DIR
    platform.upload_dir(files_dir, prefix)
    return {"status": "success"}


def files_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff files delete -p <project_directory>`
    Delete files and templates from the remote storage
    """
    state = get_state()
    state.validate_env([RESOURCE_GROUP, TASK, CLOUD_PROVIDER, CLOUD_PLATFORM,
                        BUCKET])
    platform = cloud.get_platform()
    dir_name = FILES_DIR
    platform.delete_dir(dir_name)
    return {"status": "success"}


def _config_get(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> Dict:
    """Read configs from remote parameter store and copy them to workspace dir
    """
    if not workspace_dir:
        raise Exception("Workspace directory is not set")
    LOGGER.debug("Reading configurations from remote parameter store.")
    precompiled_config = _read_project_remote(workspace_dir)

    _secrets_get(project_dir, workspace_dir, **kwargs)
    _update_state(precompiled_config, vars=vars)

    return precompiled_config


def _config_get_local(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> Dict:
    """ Compile configuration JSON file from the project.yml

    The output JSON file describes the commands and arguments for each process.
    It also contains the environment variables for the run-time.

    *.json files under config directory are encoded into the output JSON so they
    can be restored in the docker instance when it runs.

    The parameters file may contain secrets and it should kept private. (i.e.
    don't commit to a repository and etc.)

    This parameters file is encrypted and stored in a remote parameter store
    (e.g. AWS SSM) via config_push command and downloaded at the run-time.

    - project_dir: The project directory that contains:
      - project.yml
      - config:
        - *.json: JSON format configuration files necessary for each process.
        (e.g. singer tap/target config files, Google Cloud Platform secret file)

    Example project.yml:
    ```
    commands:
      - command: tap-rest-api
        args: "files/rest_api_spec.json --config config/tap_config.json --schema_dir files/schema --catalog files/catalog/default.json --state .artifacts/state --start_datetime '{start_at}' --end_datetime '{end_at}'"
        venv: "proc_01"
      - command: target_gcs
        args: "--config config/tarconfig_get.json"
        venv: "proc_02"
    envs:
      - key: "GOOGLE_APPLICATION_CREDENTIALS"
        value: "config/google_client_secret.json"
    ```
    """
    if not project_dir:
        raise Exception("Project directory is not set")
    config = dict()
    project = _read_project_local(os.path.join(project_dir, PROJECT_FILE))
    config.update(project)

    _update_state(config, vars=vars)

    return config


def config_push(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff config push -p <project_directory>`
    Push project.yml and the contents of project_dir/config as a secure
    parameter key.
    """
    platform = cloud.get_platform()
    platform.upload_file(os.path.join(project_dir, PROJECT_FILE), PROJECT_FILE)
    return {"status": "success"}


def config_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff config delete -p <project_directory>`
    Delete the project configuration from the remote parameter store.
    """
    platform = cloud.get_platform()
    platform.delete_file(PROJECT_FILE)
    return {"status": "success"}


def config_deleteold(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff config delete -p <project_directory>`
    Delete the project configuration from the remote parameter store.
    """
    platform = cloud.get_platform()
    platform.delete_parameter("config")
    return {"status": "success"}


def config_print(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff config print -p <project_directory>`
    Print the project configuration in the remote parameter store.
    """
    return _config_get(project_dir, workspace_dir, **kwargs)


def config_validate_local(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff config validate local -p <project_directory>`
    Validate local configuration.
    """
    try:
        _config_get_local(project_dir, workspace_dir)
    except ValidationError as e:
        return {"status": "error", "message": str(e)}
    return {"status": "ok", "message": "valid project.yml"}


def project_push(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff project push -p <project_directory>`
    Push config, files, secrets all together
    """
    config_push(project_dir, workspace_dir, **kwargs)
    files_push(project_dir, workspace_dir, **kwargs)
    secrets_push(project_dir, workspace_dir, **kwargs)
    return {"status": "success"}


def project_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff project push -p <project_directory>`
    Push config, files, secrets all together
    """
    config_delete(project_dir, workspace_dir, **kwargs)
    files_delete(project_dir, workspace_dir, **kwargs)
    secrets_delete(project_dir, workspace_dir, **kwargs)
    return {"status": "success"}


def workspace_init(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    if not workspace_dir:
        raise Exception("Workspace directory is not set")
    if not os.path.isdir(workspace_dir):
        os.mkdir(workspace_dir)

    artifacts_dir = os.path.join(workspace_dir, ARTIFACTS_DIR)
    files_dir = os.path.join(workspace_dir, FILES_DIR)
    if not os.path.isdir(artifacts_dir):
        os.mkdir(artifacts_dir)
    if not os.path.isdir(files_dir):
        os.mkdir(files_dir)
    return {"status": "success"}


def workspace_install(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff workspace install -p <project_directory> -w <workspace_directory>`
    Install dependencies in the local virtual environment
    """
    if not project_dir:
        raise Exception("Project directory is not set")
    if not workspace_dir:
        raise Exception("Workspace directory is not set")

    project = _read_project_local(os.path.join(project_dir, PROJECT_FILE))

    if not project.get("installs"):
        raise Exception("installs section is not defined")

    os.chdir(workspace_dir)
    for install in project["installs"]:
        if install.get("venv"):
            _make_python_venv(install["venv"])
        _install(install["command"], install.get("venv"))
    return {"status": "success"}


def version(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """Print handoff version
    """
    return {"handoff_version": VERSION}



import os, shutil
from typing import Dict

from handoff.utils import get_logger as _get_logger

LOGGER = _get_logger(__name__)


def make(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff quick_start make`
    Copy the test projects to the test_projects under the current directory
    """
    d, f = os.path.split(__file__)
    try:
        shutil.copytree(os.path.join(d, "../..",  "test_projects"),
                        os.path.join(os.getcwd(), "projects"))
    except FileExistsError:
        print("It looks like you already copied the test projects to ./projects")
    else:
        print("Copied the test projects to ./projects")
    print("Now just do:")
    print("    ./projects/start")
    print("to start the even-monkeys-can-follow tutorial.")



import argparse, datetime, json, logging, os, sys, threading, time, types
from pathlib import Path
import yaml
from lxml import html
import requests
from typing import Dict, List

from handoff.config import (VERSION, HANDOFF_DIR, ENV_PREFIX,
                            ARTIFACTS_DIR, PROJECT_FILE,
                            BUCKET, RESOURCE_GROUP, TASK,
                            CONTAINER_IMAGE, IMAGE_VERSION,
                            CLOUD_PROVIDER, CLOUD_PLATFORM,
                            CONTAINER_PROVIDER, STAGE, DEFAULT_STAGE,
                            get_state, init_state)
from handoff.core import admin, task
from handoff.utils import bcolors, get_logger
from handoff import plugins, services

LOGGER = get_logger("handoff")

LATEST_AVAILABLE_VERSION = None
ANNOUNCEMENTS = None


def _strip_outer_quotes(string: str) -> str:
    string = string.strip()
    # Avoid over stripping. Either ' or "
    if string[0] == '"':
        string = string.strip('"')
    elif string[0] == "'":
        string = string.strip("'")
    return string


def _load_param_from_file(path: str) -> Dict:
    arg_list = []
    with open(path, "r") as f:
        for l in f.readlines():
            arg_list.append(l.replace("\n", ""))
    return arg_list


def _load_param_list(arg_list: List) -> Dict:
    data = {}
    new_arg_list = []

    if len(arg_list) == 1 and os.path.isfile(arg_list[0]):
        arg_list = _load_param_from_file(arg_list[0])
    # fix the limitation in passing with $(eval echo $__VARS)
    for a in arg_list:
        if "=" in a:
            new_arg_list.append(a)
        else:
            new_arg_list[-1] = new_arg_list[-1] + " " + a

    for a in new_arg_list:
        pos = a.find("=")
        if pos < 0:
            raise ValueError("data argument list format error %s" % arg_list)
        key = _strip_outer_quotes(a[0:pos])
        value = _strip_outer_quotes(a[pos + 1:])

        if not key or not value:
            raise ValueError("data argument list format error: %s %s"
                             % (key, value))

        if value.lower() in ["true", "t"]:
            value = True
        elif value.lower() in ["false", "f"]:
            value = False
        else:
            # Convert to float or int when we can
            try:
                number = float(value)
                if "." not in value:
                    number = int(number)
                value = number
            except ValueError:
                pass

        data[key] = value

    LOGGER.debug("data params: %s" % data)
    return data


def _list_submodules(module) -> Dict:
    submodules = dict()
    objects = dir(module)
    for name in objects:
        if name[0] == "_":
            continue
        obj = getattr(module, name)

        if not callable(obj):
            submodules[name] = obj
    return submodules


def _list_commands(module: Dict) -> Dict:
    commands = dict()
    objects = dir(module)
    for name in objects:
        if name[0] == "_":
            continue
        obj = getattr(module, name)
        if not isinstance(obj, types.FunctionType):
            continue

        # Make it "command sub_command" format
        name = name.replace("_", " ")
        commands[name] = obj
    return commands


def _run_subcommand(module: types.ModuleType,
                    command: str,
                    project_dir: str,
                    workspace_dir: str,
                    show_help: bool,
                    **kwargs) -> None:
    prev_wd = os.getcwd()
    commands = _list_commands(module)
    if command in commands:
        if show_help:
            print(commands[command].__doc__ or
                  "No help doc available for %s %s" %
                  (module.__name__.split(".")[-1], command))
            os.chdir(prev_wd)
            return

        if workspace_dir:
            admin.workspace_init(project_dir, workspace_dir, **kwargs)
        response = commands[command](project_dir, workspace_dir, **kwargs)
    else:
        module_name = module.__name__.split(".")[-1]
        response = None
        print("Available commands:")
        for key in commands.keys():
            print("- " + module_name + " " + key + ": " +
                  (commands[key].__doc__ or " ").splitlines()[0])
        print("'handoff %s <command>'--help for more info" %
              module_name)
    os.chdir(prev_wd)
    return response


def _run_task_subcommand(
    command: str,
    project_dir: str,
    workspace_dir: str,
    push_artifacts: bool = False,
    debug_mode=False,
    **kwargs) -> None:
    state = get_state()
    prev_wd = os.getcwd()
    commands = _list_commands(task)

    admin.workspace_init(project_dir, workspace_dir, **kwargs)

    remote_ops = ["run", "run remote config"]
    local_ops = ["run local", "show commands"]

    # Load config from remote or local
    config = dict()
    if command in remote_ops:
        if project_dir:
            config = admin._config_get_local(
                    project_dir,
                    workspace_dir,
                    **kwargs,
                    )
        if not state.get_env(RESOURCE_GROUP) or not state.get_env(TASK):
            raise Exception(
                "To test run with remote config, you need to" +
                " either give environmental variables as:\n" +
                "  --envs resource_group=<resource_group_name>" +
                " task=<task_name>\n"  +
                " or define them at deploy in project.yml")
        config = admin._config_get(project_dir, workspace_dir, **kwargs)
        admin.files_get(project_dir, workspace_dir, **kwargs)
        admin.artifacts_get(project_dir, workspace_dir, **kwargs)
    elif command in local_ops:
        _ = admin._secrets_get_local(project_dir, workspace_dir, **kwargs)
        config = admin._config_get_local(project_dir, workspace_dir, **kwargs)
        admin.files_get_local(project_dir, workspace_dir, **kwargs)

    os.chdir(workspace_dir)

    LOGGER.debug("Running %s in %s directory" % (command, workspace_dir))
    start = datetime.datetime.utcnow()
    LOGGER.info("Job started at " + str(start))

    # Run the command
    exit_code = 0
    try:
        commands[command](config, **kwargs)
    except Exception as e:
        if debug_mode:
            raise
        LOGGER.error(str(e))
        exit_code = 1

    end = datetime.datetime.utcnow()
    LOGGER.info("Job ended at " + str(end))
    duration = end - start
    LOGGER.info("Processed in " + str(duration))

    os.chdir(prev_wd)

    push_artifacts_on_fail = config.get("deploy", {}).get(
        "push_artifacts_on_fail", True)
    if push_artifacts and (exit_code == 0 or push_artifacts_on_fail):
        if not state.get_env(BUCKET):
            raise Exception("Cannot push artifacts. BUCKET environment" +
                            "variable is not set")
        admin.artifacts_push(project_dir, workspace_dir, **kwargs)
        admin.artifacts_archive(project_dir, workspace_dir, **kwargs)


def project_required(command):
    if (not command.startswith("github") and
            not command.startswith("cloud role") and
            not command.startswith("cloud login test") and
            not command.startswith("container remote build")):
        return True
    return False


def do(
    command: str,
    project_dir: str = None,
    workspace_dir: str = None,
    show_help: bool = False,
    log_level: str = "info",
    **kwargs) -> None:
    """Determine the command to run"""
    response = {"handoff_version": VERSION}
    LOGGER.setLevel(log_level.upper())
    LOGGER.info(f"Set log level to {log_level.upper()}")
    init_state(kwargs.get("stage", os.environ.get(STAGE, DEFAULT_STAGE)))
    state = get_state()

    envs = kwargs.get("envs", {})
    for key in envs.keys():
        value = envs[key]
        if key in ["resource_group", "task"]:
            state.set_env(key.upper() + "_NAKED", value, trust=True)
            value = state["_stage-"] + value
        try:
            state.set_env(ENV_PREFIX + key.upper(), value)
        except KeyError:
            state.set_env(key.upper(), value, trust=True)
    state.set_env(
        CLOUD_PROVIDER,
        kwargs.get("cloud_provider", state.get_env(CLOUD_PROVIDER)))
    state.set_env(CLOUD_PLATFORM,
                  kwargs.get("cloud_platform", state.get_env(CLOUD_PLATFORM)))
    state.set_env(CONTAINER_PROVIDER,
                  kwargs.get("container_provider",
                             state.get_env(CONTAINER_PROVIDER)))

    # Login will cache the cloud platform credentials to state
    try:
        services.cloud.logout()
        platform = services.cloud.get_platform(vars=kwargs["vars"])
    except Exception as e:
        LOGGER.warning(str(e))

    module_name = command.split(" ")[0]
    sub_command = command[command.find(" ") + 1:]

    plugin_modules = _list_submodules(plugins)
    service_modules = _list_submodules(services)

    if command == "server":
        from handoff import ui
        vars_ = kwargs["vars"]
        ui.server(**vars_)

    # task module implements the commands such as run, run local, show commands
    if command in _list_commands(task):
        try:
            response = _run_task_subcommand(command, project_dir, workspace_dir,
                                            **kwargs)
        except Exception as e:
            raise e
        return response

    admin_commands = _list_commands(admin)
    if command == "version":
        response = admin_commands[command](project_dir, workspace_dir, **kwargs)
        return response
    if command in admin_commands:
        prev_wd = os.getcwd()
        if show_help:
            response = (admin_commands[command].__doc__ or
                        "No help doc available for " + command)
            os.chdir(prev_wd)
            return response

        try:
            if workspace_dir:
                admin.workspace_init(project_dir, workspace_dir, **kwargs)
            admin._config_get_local(project_dir, workspace_dir, **kwargs)
            r = admin_commands[command](project_dir, workspace_dir, **kwargs)
            response.update(r)
        except Exception as e:
            if log_level.lower() == "debug":
                raise
            response.update({"status": "error", "message": str(e)})

        os.chdir(prev_wd)

        # TODO: Clean up the print out logic
        print_update()
        print_announcements()
        return response

    if module_name in service_modules:
        if show_help:
            return _run_subcommand(service_modules[module_name],
                    sub_command, project_dir, workspace_dir, show_help, **kwargs)

        if project_required(command):
            admin._config_get_local(project_dir, workspace_dir, **kwargs)

        if module_name == "cloud":
            if state.get_env(CONTAINER_IMAGE) and not state.get_env(IMAGE_VERSION):
                image_version = None
                try:
                    image_version = platform.get_latest_container_image_version(
                        state.get(CONTAINER_IMAGE))
                except Exception as e:
                    LOGGER.warning("Could not get container image version: " + str(e))
                if image_version:
                    LOGGER.info(f"Container image version: {image_version}")
                    state.set_env(IMAGE_VERSION, image_version)
        try:
            r = _run_subcommand(
                    service_modules[module_name],
                    sub_command,
                    project_dir,
                    workspace_dir,
                    show_help,
                    **kwargs,
            )
            response.update(r)
        except Exception as e:
            if log_level.lower() == "debug":
                raise
            response.update({"status": "error", "message": str(e)})
        return response

    if module_name in plugin_modules.keys():
        if show_help:
            return _run_subcommand(plugin_modules[module_name], sub_command,
                                   project_dir, workspace_dir,
                                   show_help, **kwargs)

        if project_dir:
            admin._config_get_local(project_dir, workspace_dir, **kwargs)
        try:
            r = _run_subcommand(
                    plugin_modules[module_name],
                    sub_command,
                    project_dir,
                    workspace_dir,
                    show_help,
                    **kwargs,
            )
            response.update(r)
        except Exception as e:
            if log_level.lower() == "debug":
                raise
            response.update({"status": "error", "message": str(e)})
        return response

    return("Unrecognized command %s. Run handoff -h for help." % command)


def check_update() -> None:
    global LATEST_AVAILABLE_VERSION
    try:
        response = requests.get("https://pypi.org/simple/handoff")
    except Exception:
        return
    tree = html.fromstring(response.content)
    package_list = [package for package in tree.xpath("//a/text()")]
    package_list.sort(reverse=True)
    LATEST_AVAILABLE_VERSION = package_list[0][:-len(".tar.gz")]

def print_commands() -> None:
    admin_command_list = "\n- ".join(_list_commands(admin))
    task_command_list = "\n- ".join(_list_commands(task))
    plugin_list = "\n- ".join(_list_submodules(plugins).keys())

    print("""Try running:
    handoff quick_start make
in a new directory.
Check out https://dev.handoff.cloud to learn more.

Available admin commands:
- %s

Availabe run commands:
- %s

Available subcommands:
- cloud
- container
- %s

handoff <command> -h for more help.\033[0m
""" % (admin_command_list, task_command_list, plugin_list))


def print_update() -> None:
    global LATEST_AVAILABLE_VERSION
    elapsed = 0
    while LATEST_AVAILABLE_VERSION is None and elapsed < 2.0:
        time.sleep(0.25)
        elapsed += 0.25

    if (not LATEST_AVAILABLE_VERSION or
            "handoff-" + VERSION >= LATEST_AVAILABLE_VERSION):
        return

    print(bcolors.OKGREEN)
    print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
    print(":tada: handoff Ver. %s available!" % LATEST_AVAILABLE_VERSION)
    print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" + bcolors.ENDC)
    print(bcolors.BOLD + "Upgrade now:" + bcolors.ENDC +
          " pip install -U handoff\n")


def check_announcements(**kwargs) -> None:
    global ANNOUNCEMENTS
    url = "https://api.handoff.cloud/api/v1/announcements"
    try:
        response = requests.get(url, params=kwargs)
        ANNOUNCEMENTS = yaml.load(response.content, Loader=yaml.FullLoader)
        if type(ANNOUNCEMENTS) != dict:
            ANNOUNCEMENTS = None
    except Exception as e:
        LOGGER.warning(e)
    return


def print_announcements() -> None:
    global ANNOUNCEMENTS
    elapsed = 0
    while ANNOUNCEMENTS is None and elapsed < 2.0:
        time.sleep(0.25)
        elapsed += 0.25
    if not ANNOUNCEMENTS:
        return
    home = str(Path.home())
    ho_dir = os.path.join(home, HANDOFF_DIR)
    if not os.path.isdir(ho_dir):
        try:
            os.mkdir(ho_dir)
        except Exception as e:
            print("Warning: Failed to create handoff global config directory" +
                  "at %s: %s" % (ho_dir, e))
    date_file = os.path.join(ho_dir, "last_announcement_date")
    last_update = ""
    if os.path.isfile(date_file):
        with open(date_file, "r") as f:
            last_update = f.read()
    announcements = list()
    for a in ANNOUNCEMENTS.get("announcements", []):
        if a.get("date") and str(a["date"]) > last_update:
            announcements.append(a)
    if len(announcements) > 0:
        print(bcolors.OKGREEN)
        print("Announcements:" + bcolors.ENDC)
    for a in announcements:
            print(a["date"])
            print("  " + a.get("title"))
            print("  " + a.get("message"))

    try:
        with open(date_file, "w") as f:
            f.write(str(datetime.date.today()))
    except Exception as e:
        LOGGER.warning("Failed to write %s: %s" % (date_file, e))


def main() -> None:
    threading.Thread(target=check_update).start()
    parser = argparse.ArgumentParser(
        add_help=False,
        description=("handoff %s - Run parameterized Unix pipeline command." %
                     VERSION))

    parser.add_argument("command", type=str, nargs="*",
                        help="command string such as 'config push'")

    parser.add_argument("-p", "--project-dir", type=str, default=None,
                        help="Specify the location of project directory")
    parser.add_argument("-w", "--workspace-dir", type=str, default=None,
                        help="Location of workspace directory")
    parser.add_argument("-s", "--stage", type=str, default=DEFAULT_STAGE,
                        help=("Stage (default '" + DEFAULT_STAGE +
                              "'). Sets env var _stage, but keeps _stage blank when --stage 'prod'"))
    parser.add_argument("-e", "--envs", type=str, nargs="*", default="",
                        help="Define environment variables. List after this option like: -e key1=value1 key2=value2...")
    parser.add_argument("-v", "--vars", type=str, nargs="*", default="",
                        help="Extra variables for the command. List after this option like: -v key1=value1 key2=value2...")
    parser.add_argument("-y", "--yes", action="store_true", default=None,
                        help="Auto confirm with yes")
    parser.add_argument("-n", "--no", action="store_true", default=None,
                        help="Auto confirm with no")
    parser.add_argument("-d", "--debug", action="store_true", # type=bool, default=None,
                        help="Run task in debug mode")

    parser.add_argument("-h", "--help", action="store_true")
    parser.add_argument("-l", "--log-level", type=str, default="info",
                        help="Set log level (debug, info, warning, error," +
                        "critical). Default: error for non-task commands," +
                        "info for task commands")

    parser.add_argument("--cloud-provider", type=str, default="aws",
                        help="Cloud provider name")
    parser.add_argument("--cloud-platform", type=str, default="fargate",
                        help="Cloud platform name")
    parser.add_argument("--cloud-profile", type=str, default=None,
                        help="Profile name for logging in to platform")

    parser.add_argument("--container-provider", type=str, default="docker",
                        help="Container provider name")

    parser.add_argument("-a", "--push-artifacts", action="store_true",
                        help="Push artifacts to remote at the end of the run")

    if len(sys.argv) == 1 or sys.argv[1] in ["-h", "--help"]:
        parser.print_help(sys.stderr)
        print_commands()
        print_update()
        return

    args = parser.parse_args()

    if not args.command:
        parser.print_help(sys.stderr)
        print_commands()
        print_update()
        return

    args.vars = _load_param_list(args.vars)
    args.envs = _load_param_list(args.envs)

    kwargs = dict(vars(args))
    kwargs["show_help"] = args.help
    if args.yes and args.no:
        raise Exception("You cannot set yes and no at the same time")
    kwargs["yes"] = None
    if args.yes:
        kwargs["yes"] = True
    elif args.no:
        kwargs["yes"] = False
    kwargs["debug_mode"] = args.debug
    kwargs["log_level"] = args.log_level.upper()

    if (args.project_dir and args.workspace_dir and
            args.project_dir == args.workspace_dir):
        Exception("workspace directory must be different from " +
                  "project directory.")

    threading.Thread(
        target=check_announcements,
        kwargs={"command": ("-".join(args.command)).strip()}).start()

    command = (" ".join(args.command)).strip()
    kwargs.pop("command")
    response = do(command, **kwargs)

    if response:
        if isinstance(response, dict) or isinstance(response, list):
            print(yaml.dump(response))
        else:
            print(response)


if __name__ == "__main__":
    main()



#!/usr/bin/python
import io
import sys


def run():
    """Touch files except Input "4"
    """
    lines = io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8")
    for line in lines:
        line = line.strip()
        if line == "4":
            raise Exception("I don't like 4.")
        with open(f"./artifacts/out_{line}.txt", "w") as f:
            f.write("")
    print(line)


if __name__ == "__main__":
    run()



import datetime, json, logging, os, shlex, sys, venv
import subprocess
from typing import Dict

from jinja2 import Template as _Template

from handoff import utils
from handoff.config import get_state, ARTIFACTS_DIR
from handoff.core.operators import _run_task


LOGGER = utils.get_logger(__name__)


def run(config: Dict, **kwargs) -> None:
    """`handoff run -w <workspace_directory> -e resource_group=<resource_group_name> task=<task_name>`
    Run the task by the configurations and files stored in the remote parameter store and the file store.
    """
    state = get_state()
    kill_downstream_on_fail = config.get("kill_downstream_on_fail", True)
    kill_loop_on_fail = config.get("kill_loop_on_fail", True)
    exit_code = 0
    for task in config["tasks"]:
        if not task.get("active", True):
            continue
        kill_ds_on_fail = task.get(
            "kill_downstream_on_fail",
            kill_downstream_on_fail,
        )
        kill_lp_on_fail = task.get(
            "kill_loop_on_fail",
            kill_loop_on_fail,
        )
        try:
            exit_code = _run_task(
                task,
                state,
                ARTIFACTS_DIR,
                kill_downstream_on_fail=kill_ds_on_fail,
                kill_loop_on_fail=kill_lp_on_fail,
            )
        except Exception as e:
            LOGGER.error(str(e))
            exit_code = 1
        LOGGER.info("Task %s exited with code %d" %
                    (task.get("name", ""), exit_code))
        if exit_code > 0:
            if kill_downstream_on_fail:
                LOGGER.warning("Not running the rest of the tasks.")
                break
    return exit_code


def run_local(config, **kwargs):
    """`handoff run local -p <project_directory> -w <workspace_directory>`
    Run the task locally.
    """
    return run(config, **kwargs)



import logging
import boto3
from . import credentials as cred

logger = logging.getLogger(__name__)


def get_client(cred_keys: dict = {}):
    return cred.get_client("sts", cred_keys)


def get_default_region(cred_keys: dict = {}):
    client = cred.get_session(cred_keys)
    return client.region_name


def get_account_id(cred_keys: dict = {}):
    client = get_client(cred_keys)
    account_id = client.get_caller_identity()["Account"]
    return account_id


def assume_role(
        role_arn: str,
        session_name: str = "tmp",
        external_id: str = None,
        cred_keys: dict = {}):
    client = get_client(cred_keys)
    kwargs = {
        "RoleArn": role_arn,
        "RoleSessionName": session_name,
    }
    if external_id:
        kwargs["ExternalId"] = external_id
    return client.assume_role(**kwargs)



#!/usr/bin/python
import io, json, logging, sys, os

LOGGER = logging.getLogger()

def run():
    """Create long format
    """
    schema = {"type": "SCHEMA", "stream": "exchange_rate_long_format",
              "stream": "exchange_rate_long_format",
              "schema": {"type": "object",
                         "properties": {
                             "date": {"type": "string", "format": "date-time"},
                             "symbol": {"type": ["null", "string"]},
                             "rate": {"type": ["null", "number"]}}},
              "key_properties": ["date"]
              }
    print(json.dumps(schema))
    lines = io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8")
    for line in lines:
        o = json.loads(line)
        if o["type"].lower() != "record":
            if o["type"].lower() != "schema":
                print(json.dumps(o))
            continue
        record = o["record"]
        date = record["date"]
        for key in record.keys():
            if key == "date":
                continue
            r = {"type": "RECORD",
                 "stream": "exchange_rate_long_format",
                 "record": {"date": date, "symbol": key, "rate": record[key]}}
            print(json.dumps(r))


if __name__ == "__main__":
    run()



import os, shutil, sys, tarfile
import yaml
from importlib import import_module as _import_module
from typing import Dict
from types import ModuleType

from handoff.core import admin
from handoff import utils
from handoff.utils import get_logger as _get_logger
from handoff.config import (BUCKET, RESOURCE_GROUP, TASK, CONTAINER_IMAGE,
                            IMAGE_VERSION, CLOUD_PROVIDER, CLOUD_PLATFORM,
                            STAGE, TASK_NAKED)
from handoff.config import get_state as _get_state


LOGGER = _get_logger(__name__)


PLATFORM_MODULE = None


def logout():
    global PLATFORM_MODULE
    PLATFORM_MODULE = None


def get_platform(
    provider_name: str = None,
    platform_name: str = None,
    stdout: bool = False,
    cloud_profile: str = None,
    vars: Dict = {},
    **kwargs) -> ModuleType:
    global PLATFORM_MODULE
    if PLATFORM_MODULE:
        return PLATFORM_MODULE

    state = _get_state()

    if not provider_name:
        provider_name = state.get(CLOUD_PROVIDER)
    if not platform_name:
        platform_name = state.get(CLOUD_PLATFORM)
    if not provider_name or not platform_name:
        raise Exception("You need to set provider_name and platform_name")

    PLATFORM_MODULE = _import_module("handoff.services.cloud."
                                     + provider_name)
    cred_keys = PLATFORM_MODULE.find_cred_keys(vars)
    response = PLATFORM_MODULE.login(cloud_profile, cred_keys=cred_keys)
    if not response:
        raise Exception(
            f"Login to {provider_name} failed. Credentials may not be set correctly.")

    PLATFORM_MODULE.set_log_level()
    return PLATFORM_MODULE


def login_test(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    platform = get_platform()
    account_id = platform.get_account_id()
    if not account_id:
        return {"status": "not connected", "account_id": account_id}
    return {"status": "success", "account_id": account_id}


def role_create(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud role create -p <project_directory> -v external_id=<id> grantee_account_id=<grantee_id>`
    Create the role with deployment privilege.
    """
    state = _get_state()
    platform = get_platform()
    account_id = platform.get_account_id()
    if not account_id:
        raise Exception("Login failed")
    # state.validate_env()
    if not vars.get("grantee_account_id"):
        LOGGER.warn("grantee_account_id was not set." +
                    "The grantee will be set for the same account. To set: ")
        LOGGER.warn("-v grantee_account_id=xxxx")
    if not vars.get("external_id"):
        raise ValueError("external_id must be set. Do as:\n    " +
                         "handoff cloud create_role -p <project-vir> " +
                         "-v external_id=yyyy")

    return platform.create_role(
        grantee_account_id=str(vars.get("grantee_account_id", account_id)),
        external_id=vars.get("external_id"),
        )


def role_update(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud role update -p <project_directory> -v external_id=<id> grantee_account_id=<grantee_id>`
    Update the role privilege information.
    """
    state = _get_state()
    platform = get_platform()
    account_id = platform.get_account_id()
    if not account_id:
        raise Exception("Login failed")
    # state.validate_env()
    if not vars.get("grantee_account_id"):
        LOGGER.warn("grantee_account_id was not set." +
                    "The grantee will be set for the same account. To set: ")
        LOGGER.warn("-v grantee_account_id=xxxx")
    if not vars.get("external_id"):
        raise ValueError("external_id must be set. Do as:\n    " +
                         "handoff cloud create_role -p <project-vir> " +
                         "-v external_id=yyyy")

    return platform.update_role(
        grantee_account_id=str(vars.get("grantee_account_id", account_id)),
        external_id=vars.get("external_id")
    )


def role_delete(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud role delete -p <project_directory> -v grantee_account_id=<grantee_id>`
    Delete the role.
    """
    state = _get_state()
    platform = get_platform()
    account_id = platform.get_account_id()
    # state.validate_env()
    if not vars.get("grantee_account_id"):
        LOGGER.warn("grantee_account_id was not set." +
                    "The grantee will be set for the same account. To set: ")
        LOGGER.warn("-v grantee_account_id=xxxx")

    return platform.delete_role(
        grantee_account_id=str(vars.get("grantee_account_id", account_id))
    )


def role_status(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud role status -p <project_directory>` -s <stage>
    Check the status of role
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.get_role_status(
        grantee_account_id=str(vars.get("grantee_account_id", account_id))
    )


def bucket_create(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff cloud bucket create -p <project_directory>`
    Create remote storage bucket. Bucket is attached to the resource group.
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.create_bucket()


def bucket_update(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff cloud bucket update -p <project_directory>`
    Update remote storage bucket info
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.update_bucket()


def bucket_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff cloud bucket delete -p <project_directory>`
    Delete remote storage bucket.
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.delete_bucket()


def bucket_status(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud bucket status -p <project_directory>` -s <stage>
    Check the status of bucket
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.get_bucket_status(**vars)


def resources_create(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud resources create -p <project_directory>`
    Create resources necessary for task execution.
    The resources are shared among the tasks under the same resource group.

    Optionally:
    -v static_ip=True: [AWS] Create Elastic IP and NAT Gateway to obtain a static IP. This [costs more](https://aws.amazon.com/vpc/pricing/).

    AWS:
    - Please refer to the
      [CloudFormation template](https://github.com/anelenvars/handoff/blob/master/handoff/services/cloud/aws/cloudformation_templates/resources.yml) for the resources created with this command.
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.create_resources(**vars)


def resources_update(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud resources update -p <project_directory>`
    Update the resources

    The resources are shared among the tasks under the same resource group.
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.update_resources(**vars)


def resources_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff cloud resources delete -p <project_directory>`
    Delete the resources

    The resources are shared among the tasks under the same resource group.
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.delete_resources()


def resources_status(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud resources status -p <project_directory>` -s <stage>
    Check the status of resources
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.get_resources_status(**vars)


def task_create(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud task create -p <project_directory>`
    Create the task
    Optionally,
    -v cpu=256, memory=512 storage=21
    The above example is for AWS Fargate with vCPU 256 unit, 512MB memory, and 21GB ephemeral storage
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env([IMAGE_VERSION])
    return platform.create_task(**vars)


def task_update(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud task update -p <project_directory>`
    Update the task
    Optionally,
    -v cpu=256, memory=512
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env([IMAGE_VERSION])
    return platform.update_task(**vars)


def task_delete(
    project_dir: str,
    workspace_dir: str,
    **kwargs) -> None:
    """`handoff cloud task delete -p <project_directory>`
    Delete the task
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.delete_task()


def task_status(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud task status -p <project_directory>` -s <stage>
    Check the status of task
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.get_task_status(**vars)


def job_status(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud job status -p <project_directory> -v full=False running=True stopped=True resource_group_level=False`
    list task status

    AWS options:
    - full: When true, print the full description (default: false)
    - running: When true, include desired status = RUNNING (default: true)
    - stopped: When true, include desired status = STOPPED (default: true)
    - resource_group_level: When true, list all the tasks under the same resource groups (default: false)
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    ret = platform.list_jobs(**vars)
    return ret


def job_stop(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud job stop -p <project_directory> -v id=<task_id> reason=<reason>`
    stop a running task
    Options:
    - reason
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    return platform.stop_job(**vars)


def run(
    project_dir: str,
    workspace_dir: str,
    envs: Dict = {},
    vars: Dict = {},
    extras: str = None,
    **kwargs) -> None:
    """`handoff cloud run -v resource_group=<resource_group_name> task=<task_name> target_id=<target_id> -e vars='key1=val1 key2=val2...'`
    Run a task once in the platform

    If the environment variable vars is specified via -e option, it will be
    used as:
    `handoff run -v $(eval echo $vars)`
    """
    state = _get_state()
    platform = get_platform()
    config = admin._config_get_local(project_dir, workspace_dir)
    state.validate_env()

    extras_obj = None
    if extras:
        with open(extras, "r") as f:
            extras_obj = yaml.load(f)

    target_id = vars.get("target_id")
    target_envs = dict()
    if target_id:
        for s in config.get("schedules", []):
            if str(s["target_id"]) == str(target_id):
                for e in s.get("envs", []):
                    target_envs[e["key"]] = e["value"]
                break
    target_envs.update(envs)
    target_envs[STAGE] = state[STAGE]
    return platform.run_job(env=target_envs, extras=extras_obj)


def container_build(
    project_dir: str,
    workspace_dir: str,
    envs: Dict = {},
    vars: Dict = {},
    extras: str = None,
    yes: bool = False,
    **kwargs) -> None:
    """`handoff cloud container build -v resource_group=<resource_group_name> task=<task_name> target_id=<target_id> -e vars='key1=val1 key2=val2...'`
    build and push container in the cloud
    """
    state = _get_state()
    platform = get_platform()
    config = admin._config_get_local(project_dir, workspace_dir)
    state.validate_env()

    # Make the image repository if it does not exist
    username, password, registry = platform.get_docker_registry_credentials()
    image_name = state.get(CONTAINER_IMAGE)
    try:
        platform.get_repository_images(image_name)
    except Exception:
        if not yes:
            sys.stdout.write("Repository %s does not exist. Create (y/N)?" %
                             image_name)
            response = input()
            if response.lower() not in ["y", "yes"]:
                return
        LOGGER.info("Creating repository " + image_name)
        platform.create_repository()

    tar_file_name = state[TASK_NAKED] + ".tar.gz"
    account_id = platform.get_account_id()
    version = platform.get_latest_container_image_version(state[CONTAINER_IMAGE])
    new_version = utils.increment_version(version)
    destination = f"{account_id}.dkr.ecr.{state['AWS_REGION']}.amazonaws.com/{state[CONTAINER_IMAGE]}:{new_version}"
    command = [
        "--context", f"s3://{state[BUCKET]}/{state[TASK_NAKED]}/bundles/{tar_file_name}",
        #"--context-sub-path", "./some_subdir",
        #"--dockerfile", "myDockerfile",
        "--destination", destination,
        "--force",
    ]

    extras_obj = None
    if extras:
        with open(extras, "r") as f:
            extras_obj = yaml.load(f)

    target_envs = {}
    target_envs.update(envs)
    target_envs[STAGE] = state[STAGE]
    target_envs["COMMAND"] = "container remote build"
    target_envs[TASK] = state[TASK_NAKED]

    return platform.run_job(
            task_name="handoff-container-builder",
            container_name="handoff-container-builder",
            env=target_envs,
            command=command,
            extras=extras_obj,
            )


def container_version(
    project_dir: str,
    workspace_dir: str,
    envs: Dict = {},
    vars: Dict = {},
    extras: str = None,
    yes: bool = False,
    **kwargs) -> None:
    state = _get_state()
    platform = get_platform()
    image_name = state.get(CONTAINER_IMAGE)
    try:
        version = platform.get_latest_container_image_version(state[CONTAINER_IMAGE])
    except Exception as e:
        return {"status": "error", "message": str(e)}

    return {"status": "success", "version": version}


def schedule_create(
    project_dir: str,
    workspace_dir: str,
    envs: Dict = {},
    vars: Dict = {},
    extras: str = None,
    **kwargs) -> None:
    """`handoff cloud schedule create -v target_id=<target_id> cron="<cron_format>" -e vars='key1=val1 key2=val2...'`
    Schedule a task named <target_id> at the recurring scheduled specified
    as <cron_format>. An example of cron-format string is `10 01 * * ? *`
    for every day at 01:10 (1:10AM)

    If the environment variable vars is specified via -e option, it will be
    used as:
    `handoff run -v $(eval echo $vars)`
    """
    state = _get_state()
    platform = get_platform()
    config = admin._config_get_local(project_dir, workspace_dir)
    state.validate_env()
    schedules = config.get("schedules")

    platform_config = config.get("deploy", {}).get("platform_config", {})
    kwargs = {}
    kwargs.update(platform_config)

    target_id = vars.get("target_id")
    cronexp = vars.get("cron")
    extras_obj = None
    if extras:
        with open(extras, "r") as f:
            extras_obj = yaml.load(f)
    if not schedules:
        schedules = [{"target_id": target_id, "cron": cronexp,
                      "extras_obj": extras_obj}]

    if not schedules and (not target_id or not cronexp):
        return {
                "status": "error",
                "message": "Schedules not found in project.yml. You can also set "
                           "'-v target_id=<ID> cron=<CRON>'",
              }

    envs[STAGE] = state[STAGE]

    responses = []
    for s in schedules:
        if target_id is not None and str(s["target_id"]) != str(target_id):
            continue
        if target_id is not None and cronexp:
            s["cron"] = cronexp

        e = dict()
        for e1 in s.get("envs", list()):
            e[e1["key"]] = e1["value"]
        # priority of the values schedule section < envs section < command line
        e.update(envs)
        r = platform.schedule_job(
            str(s["target_id"]), "cron(" + s["cron"] + ")",
            env=e,
            extras=s.get("extras_obj"),
            **kwargs)
        responses.append(r)
    return {"status": "success", "result": responses}



def schedule_delete(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud schedule delete -v target_id=<target_id>`
    Unschedule a task named <target_id>
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    if not vars.get("target_id"):
        return {
            "status": "error",
            "message": "Forgot to set '-v target_id=<ID>' ?",
        }
    target_id = str(vars["target_id"])
    try:
        response = platform.unschedule_job(target_id)
    except Exception as e:
        response = str(e)
    return {"status": "success", "result": responses}


def schedule_list(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud schedule list`
    List the scheduled tasks
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    schedules = platform.list_schedules(**vars)
    target_ids = []
    for s in schedules["schedules"]:
        target_ids.append(str(s["target_id"]))
        s["status"] = "scheduled"

    if vars.get("include_unpublished"):
        config = admin._config_get_local(project_dir, workspace_dir)
        local_schedules = config.get("schedules", [])
        for s in local_schedules:
            s["target_id"] = str(s["target_id"])
            if s.get("envs") is None:
                s["envs"] = []
            status = ""
            try:
                index = target_ids.index(str(s["target_id"]))
            except ValueError:
                index = len(target_ids)
                target_ids.append(s["target_id"])
                schedules["schedules"].append({})
                status = "draft"

            remote = {}
            remote.update(schedules["schedules"][index])
            if remote.get("envs") is None:
                remote["envs"] = []
            if remote.get("status"):
                remote.pop("status")
            if remote.get("name"):
                remote.pop("name")
            if s.get("description"):
                s.pop("description")
            if s != remote:
                LOGGER.debug(f"local : {str(s)}")
                LOGGER.debug(f"remote: {str(remote)}")
                schedules["schedules"][index] = s
                status = status or "edited"
                schedules["schedules"][index]["status"] = status

    return schedules


def logs(
    project_dir: str,
    workspace_dir: str,
    vars: Dict = {},
    **kwargs) -> None:
    """`handoff cloud logs -v start_time=<start_time> end_time=<end_time> follow=<True/False>`
    Show logs
    Use --vars (-v) option to:
    - start_time: ISO 8086 formatted date time to indicate the start time
    - end_time
    - follow: If set, it waits for more logs until interrupted by ctrl-c
    - filter: Filter log term
    """
    state = _get_state()
    platform = get_platform()
    state.validate_env()
    close = False
    if vars.get("file"):
        file_descriptor = open(vars.pop("file"), "a")
        close = True
    else:
        file_descriptor = sys.stdout

    if vars.get("role_arn"):
        vars.pop("role_arn")
    platform.write_logs(file_descriptor, **vars)

    if close:
        file_descriptor.close()



#!/usr/bin/env python
from setuptools import setup

VERSION = "0.4.0"

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="handoff",
    version=VERSION,
    description="Deploy configurable unix pipeline jobs serverlessly.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Daigo Tanaka, Anelen Co., LLC",
    url="http://dev.handoff.cloud",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",

        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",

        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        "attrs>=17.4.0",
        "boto3>=1.10.0",
        "botocore>=1.13.0",
        "python-dateutil>=2.7.0",
        "docker==5.0.0",
        "lxml>=4.2.0",
        "Jinja2>=2.10.1",
        "jsonschema>=3.2.0",
        "packaging>=19.2",
        "pygit2==1.5.0",
        "PyGithub>=1.50",
        "python-dateutil>=2.8.0",
        "pyyaml>=5.4",
        "requests>=2.19.0",
        "s3transfer>=0.2.0",
    ],
    entry_points="""
    [console_scripts]
    handoff=handoff:main
    """,
    packages=["handoff"],
    package_data={
        # Use MANIFEST.ini
    },
    include_package_data=True
)



import logging

import boto3

from . import credentials as cred


logger = logging.getLogger(__name__)


def get_client(cred_keys: dict = {}):
    return cred.get_client("iam", cred_keys)


def list_roles(path_prefix=None, cred_keys: dict = {}):
    client = get_client(cred_keys)
    kwargs = dict()
    if path_prefix:
        kwargs["PathPrefix"] = path_prefix
    response = client.list_roles(**kwargs)
    roles = response["Roles"]
    marker = None
    if response["IsTruncated"]:
        marker = response["Marker"]
    while marker:
        kwargs["Marker"] = marker
        marker = None
        response = client.list_roles(**kwargs)
        roles = roles + response["Roles"]
        if response["IsTruncated"]:
            marker = response["Marker"]
    return roles



import json, logging, os

import boto3

from . import credentials as cred
from . import cloudformation as cfn, iam, sts, stepfunctions as stepfns

from handoff import utils


LOGGER = utils.get_logger()


def get_client(cred_keys: dict = {}):
    return cred.get_client("events", cred_keys)


def schedule_job(
        account_id,
        task_stack,
        resource_group_stack,
        container_image,
        region,
        target_id,
        cronexp,
        role_arn,
        env=[],
        state_machine=True,
        state_machine_timeout=3600 * 4,
        extras=None,
        cred_keys: dict = {},
        **kwargs):
    """Schedule a task
    extras overwrite the kwargs given to put_targets boto3 command.
    See: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/events.html#EventBridge.Client.put_targets
    """
    client = get_client(cred_keys)

    rule_name = task_stack + "-" + target_id

    rule_kwargs = {
        "Name": rule_name,
        "ScheduleExpression": cronexp
    }
    client.put_rule(**rule_kwargs)

    task_resources = cfn.describe_stack_resources(
            task_stack,
            cred_keys=cred_keys,
        )["StackResources"]
    task_def_arn = None
    for r in task_resources:
        if r["ResourceType"] == "AWS::ECS::TaskDefinition":
            task_def_arn = r["PhysicalResourceId"]
            break

    rg_resources = cfn.describe_stack_resources(
            resource_group_stack,
            cred_keys=cred_keys,
        )["StackResources"]
    cluster_arn = None
    subnets = list()
    security_groups = list()
    for r in rg_resources:
        if not cluster_arn and r["ResourceType"] == "AWS::ECS::Cluster":
            phys_rsrc_id= r["PhysicalResourceId"]
            cluster_arn = f"arn:aws:ecs:{region}:{account_id}:cluster/{phys_rsrc_id}"
        # PublicSubnetTwo is <= v0.3.4
        elif (r["ResourceType"] == "AWS::EC2::Subnet" and
                r["LogicalResourceId"] in ["PublicSubnetTwo", "Subnet2"]):
            subnets.append(r["PhysicalResourceId"])
        elif r["ResourceType"] == "AWS::EC2::SecurityGroup":
            security_groups.append(r["PhysicalResourceId"])

    task_resources = cfn.describe_stack_resources(
            task_stack,
            cred_keys=cred_keys,
        )["StackResources"]
    log_group_arn = None
    for r in task_resources:
        if r["ResourceType"] == "AWS::Logs::LogGroup":
            phys_rsrc_id = r["PhysicalResourceId"]
            log_group_arn = f"arn:aws:logs:{region}:{account_id}:log-group:{phys_rsrc_id}:*"

    LOGGER.debug(f"{cluster_arn}\n{task_def_arn}\n{subnets}\n{security_groups}\n{log_group_arn}")

    overrides = {
        "ContainerOverrides": [{
            "Name": container_image,
            "Environment": env,
        }]
    }

    if state_machine:
        sm_arn = stepfns.create_state_machine(
            rule_name,
            cluster_arn,
            task_def_arn,
            subnets,
            security_groups,
            role_arn,
            overrides=overrides,
            log_group_arn=log_group_arn,
            timeout=state_machine_timeout,
            cred_keys=cred_keys,
            **kwargs,
        )
        target = {
            "Id": target_id,
            "Arn": sm_arn,
            "RoleArn": role_arn,
            "Input": json.dumps(overrides),
        }
    else:
        target = {
            "Id": target_id,
            "Arn": cluster_arn,
            "RoleArn": role_arn,
            "EcsParameters": {
                "TaskDefinitionArn": task_def_arn,
                "TaskCount": 1,
                "LaunchType": "FARGATE",
                "NetworkConfiguration": {
                    "awsvpcConfiguration": {
                        "Subnets": subnets,
                        "SecurityGroups": security_groups,
                        "AssignPublicIp": "ENABLED"
                    }
                }
            },
            "Input": json.dumps(overrides)
        }

    rule_kwargs = {
        "Rule": rule_name,
        "Targets": [target]
    }

    if extras:
        rule_kwargs.update(extras)

    return client.put_targets(**rule_kwargs)


def unschedule_job(
        task_stack,
        target_id,
        cred_keys: dict = {}):
    client = get_client(cred_keys)
    rule_name = task_stack + "-" + target_id
    targets = client.list_targets_by_rule(Rule=rule_name)
    target_ids = []
    for target in targets["Targets"]:
        target_ids.append(target["Id"])
        if "stateMachine" in target["Arn"]:
            LOGGER.debug(f"Removing stepfunction {target['Arn']}")
            stepfns.delete_state_machine(
                target["Arn"],
                cred_keys=cred_keys,
            )

    kwargs = {
        "Rule": rule_name,
        "Ids": target_ids
    }
    client.remove_targets(**kwargs)
    client.delete_rule(Name=rule_name)


def list_schedules(
        task_stack,
        cred_keys: dict = {},
        ):
    client = get_client(cred_keys)
    rule_prefix = task_stack
    kwargs = {
        "NamePrefix": rule_prefix,
    }
    rules = client.list_rules(**kwargs)
    if not rules.get("Rules"):
        raise Exception("No rules found")

    response = list()
    for rule in rules.get("Rules"):
        targets = client.list_targets_by_rule(Rule=rule["Name"])
        record = {"rule": rule, "targets": targets.get("Targets")}
        response.append(record)
    return response
